"""Tools for probing, querying, inspecting and drawing CompartmentalModels

The main entry point for this is the ModelProbe class
"""

import re
from typing import Iterable, Set, List

import networkx as nx
import numpy as np

from summer2 import CompartmentalModel
from summer2.compartment import Compartment
from summer2.flows import BaseFlow


def query_compartments(m: CompartmentalModel, query: dict = None, tags: List = None, as_idx=False):
    query = query or {}
    tags = tags or []
    if isinstance(tags, str):
        tags = [tags]
    if "name" in query:
        query = query.copy()
        name = query.pop("name")
        matching = m.get_matching_compartments(name, query)
        if as_idx:
            return np.array(
                [c.idx for c in matching if all([t in c.tags for t in tags])], dtype=int
            )
        else:
            return [c for c in matching if all([t in c.tags for t in tags])]
    else:
        _strata = frozenset(query.items())
        if as_idx:
            return np.array(
                [
                    c.idx
                    for c in m.compartments
                    if c._has_strata(_strata) and all([t in c.tags for t in tags])
                ],
                dtype=int,
            )
        else:
            return [
                c
                for c in m.compartments
                if c._has_strata(_strata) and all([t in c.tags for t in tags])
            ]


def query_flows(
    m: CompartmentalModel,
    flow_name: str = None,
    source: dict = None,
    dest: dict = None,
    tags: List = None,
):
    if flow_name is not None:
        if isinstance(flow_name, re.Pattern):
            flows = [f for f in m._flows if flow_name.match(f.name)]
        elif isinstance(flow_name, str):
            flows = [f for f in m._flows if flow_name == f.name]
        else:
            flows = flow_name
    else:
        flows = m._flows

    if source:
        if "name" in source:
            source = source.copy()
            name = source.pop("name")
            flows = [f for f in flows if f.source and f.source.name == name]
        else:
            source = frozenset(source.items())
            flows = [f for f in flows if f.source and f.source._has_strata(source)]

    if dest:
        if "name" in dest:
            dest = dest.copy()
            name = dest.pop("name")
            flows = [f for f in flows if f.dest and f.dest.name == name]
        else:
            dest = frozenset(dest.items())
            flows = [f for f in flows if f.dest and f.dest._has_strata(source)]

    if tags:
        if isinstance(tags, str):
            tags = [tags]
        flows = [f for f in flows if all([t in f.tags for t in tags])]

    return flows


def flows_to_compartments(m: CompartmentalModel, flows: Iterable[BaseFlow]):
    comps = []
    for f in flows:
        if f.source:
            comps.append(f.source)
        if f.dest:
            comps.append(f.dest)
    return set(comps)


def build_compartment_flow_map(m: CompartmentalModel):
    out_map = {c: set() for c in m.compartments}
    for f in m._flows:
        if f.source:
            out_map[f.source].add(f)
        if f.dest:
            out_map[f.dest].add(f)
    return out_map


class ModelProbe:
    def __init__(self, model: CompartmentalModel):
        self.model = model
        self._compartment_flow_map = build_compartment_flow_map(model)

    def compartments_to_flows(self, compartments: Iterable[Compartment]) -> Set[BaseFlow]:
        flows = []
        for c in compartments:
            flows += self._compartment_flow_map[c]
        return set(flows)

    def flows_to_compartments(self, flows: Iterable[BaseFlow]) -> Set[Compartment]:
        return flows_to_compartments(self.model, flows)

    def query_compartments(self, query: dict = None, tags: List = None, as_idx=False):
        return query_compartments(self.model, query, tags, as_idx)

    def query_flows(
        self, flow_name: str = None, source: dict = None, dest: dict = None, tags: List = None
    ):
        return query_flows(self.model, flow_name, source, dest, tags)

    def get_model_subset(self, comp_query: dict = None, flow_query: dict = None):

        comp_query = comp_query or {}
        flow_query = flow_query or {}

        comps = self.query_compartments(comp_query)
        flows = self.query_flows(**flow_query)

        matched_comps = self.flows_to_compartments(flows)
        matched_flows = self.compartments_to_flows(comps)

        return matched_comps.intersection(set(comps)), matched_flows.intersection(set(flows))

    def draw_flow_graph(self, comp_query: dict = None, flow_query: dict = None):

        # FIXME: Not yet working - need to update with new computegraph methods

        raise NotImplementedError()
        # comp_query = comp_query or {}
        # flow_query = flow_query or {}

        # comps, flows = self.get_model_subset(comp_query, flow_query)
        # digraph = model_to_digraph(comps, flows)

        # pos = nx.nx_agraph.graphviz_layout(digraph)

        # edge_trace, node_trace = ngraph.get_traces(digraph, pos)

        # node_trace.marker.color = ['#ff3333' if 'infectious' in str(node) else \
        # '#dd44dd' if 'latent' in str(node) else '#22dd22' for node in digraph.nodes]
        # node_text = [
        #   get_node_text_summer(probe.model, n, node) for n,node in digraph.nodes.items()
        # ]
        # node_trace.text = node_text

        # title = 'Compartmental Model'

        # return get_graph_figure(edge_trace, node_trace, title)


"""
Tools specific to handling models as networkx graph structures
"""


def model_to_digraph(compartments: Iterable[Compartment], flows: Iterable[BaseFlow]):
    g = nx.DiGraph()
    for c in compartments:
        g.add_node(c)

    def is_fully_mapped(f, comps):
        return f.source in comps and f.dest in comps

    for f in flows:
        if is_fully_mapped(f, compartments):
            g.add_edge(str(f.source), str(f.dest))

    return g



"""
Calculation of derived outputs.

A derived output is an additional output that the user has requested,
which can be calculated (or "derived") using the model results, which are the:

    - evaluation times
    - compartment sizes for each time
    - flow rates for each time

"""
import logging

from jax import numpy as jnp
import numpy as np

from computegraph import ComputeGraph
from computegraph.utils import relabel_tree
from computegraph.types import local

from summer2.parameters import Function, ModelVariable

logger = logging.getLogger()


def build_flow_output(request, name, times, model_flows, idx_cache=None):

    flow_indices = []
    for flow_idx, flow in enumerate(model_flows):
        is_matching_flow = (
            flow.name == request["flow_name"]
            and ((not flow.source) or flow.source.has_strata(request["source_strata"]))
            and ((not flow.dest) or flow.dest.has_strata(request["dest_strata"]))
        )
        if is_matching_flow:
            flow_indices.append(flow_idx)

    flow_indices = jnp.array(flow_indices)

    use_raw_results = request["raw_results"]

    if use_raw_results:

        def get_flow_output(flows):
            return flows[:, flow_indices].sum(axis=1)

    else:

        def get_flow_output(flows):
            flow_vals = flows[:, flow_indices].sum(axis=1)
            midpoint_output = jnp.zeros(times.shape)
            midpoint_output = midpoint_output.at[0].set(flow_vals[0])
            interp_vals = (flow_vals[1:] + flow_vals[:-1]) * 0.5
            midpoint_output = midpoint_output.at[1:].set(interp_vals)
            return midpoint_output

    return Function(get_flow_output, [ModelVariable("flows")])


def build_compartment_output(request, name, compartments):

    req_compartments = request["compartments"]
    strata = request["strata"]
    comps = ((i, c) for i, c in enumerate(compartments) if c.has_name_in_list(req_compartments))
    indices = [i for i, c in comps if c.is_match(c.name, strata)]

    def summed_compartment_outputs(outputs):
        return outputs[:, indices].sum(axis=1)

    return Function(summed_compartment_outputs, [ModelVariable("outputs")])


def return_agg(*sources):
    return jnp.array(sources).sum(axis=0)


def build_aggregate_output(request):
    return Function(return_agg, [local(src) for src in request["sources"]])


def build_cumulative_output(request, name, times, baseline_offset=None):
    source_name = request["source"]
    start_time = request["start_time"]
    max_time = times.max()
    if start_time and start_time > max_time:
        # Handle case where the derived output starts accumulating after the last model timestep.
        msg = f"Cumulative output '{name}' start time {start_time} is greater than max model time {max_time}, defaulting to {max_time}"
        logger.warn(msg)
        start_time = max_time

    if baseline_offset is not None:
        raise NotImplementedError()

    if start_time is None:
        return Function(jnp.cumsum, [local(source_name)])
    else:
        assert start_time in times, f"Start time {start_time} not in times for '{name}'"
        start_idx = np.where(times == start_time)[0][0]

        def get_indexed_cumsum(in_arr):
            output = jnp.zeros(len(times), dtype=jnp.float64)
            output = output.at[start_idx:].set(jnp.cumsum(in_arr[start_idx:]))
            return output

        return Function(get_indexed_cumsum, [local(source_name)])


def build_function_output(request):
    func = request["func"]
    source_names = request["sources"]
    inputs = [local(s) for s in source_names]
    return Function(func, inputs)


def build_computed_value_output(request, name):
    return Function(lambda x: x[name], [ModelVariable("computed_values")])


def build_derived_outputs_runner(model):
    graph_dict = {}
    out_keys = []
    for name, request in model._derived_output_requests.items():
        req_type = request["request_type"]
        if req_type == "comp":
            graph_dict[name] = build_compartment_output(request, name, model.compartments)
        elif req_type == "param_func":
            graph_dict[name] = relabel_tree(request["func"], "derived_outputs", "graph_locals")
        elif req_type == "flow":
            graph_dict[name] = build_flow_output(request, name, model.times, model._flows)
        elif req_type == "agg":
            graph_dict[name] = build_aggregate_output(request)
        elif req_type == "cum":
            graph_dict[name] = build_cumulative_output(request, name, model.times)
        elif req_type == "func":
            graph_dict[name] = build_function_output(request)
        elif req_type == "computed_value":
            graph_dict[name] = build_computed_value_output(request, name)
        else:
            raise NotImplementedError(request)
        if request["save_results"]:
            out_keys.append(name)

    if model._derived_outputs_whitelist:
        out_keys = model._derived_outputs_whitelist

    cg = ComputeGraph(graph_dict)
    return cg, cg.get_callable(targets=out_keys)



from computegraph import jaxify as _jaxify

_jaxify.set_using_jax(True)

from .adjust import Multiply, Overwrite
from .compartment import Compartment
from .model import CompartmentalModel
from .stratification import AgeStratification, StrainStratification, Stratification



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from computegraph.types import Variable, Function, GraphObject, Data, build_args  # noqa: F401
from computegraph.utils import extract_variables, is_var

if TYPE_CHECKING:
    from summer2 import CompartmentalModel


class Parameter(Variable):
    def __init__(self, key: str):
        super().__init__(key, "parameters")

    def __repr__(self):
        return f"Parameter {self.key}"


class DerivedOutput(Variable):
    def __init__(self, name: str):
        super().__init__(name, "derived_outputs")

    def __repr__(self):
        return f"DerivedOutput {self.key}"


class ModelVariable(Variable):
    def __init__(self, name: str):
        super().__init__(name, "model_variables")

    def __repr__(self):
        return f"ModelVariable {self.key}"


CompartmentValues = ModelVariable("compartment_values")
ComputedValuesDict = Variable("computed_values", "graph_locals")
Time = ModelVariable("time")


def is_func(param) -> bool:
    """Wrapper to handle Function or callable

    Args:
        param: Flow or adjustment parameter

    Returns:
        bool: Is a function or function wrapper
    """
    from .param_impl import GraphObjectParameter

    return isinstance(param, Function) or callable(param)


def get_model_param_value(
    param: Any, time: float, computed_values: dict, parameters: dict, mul_outputs=False
) -> Any:
    """Get the value of anything that might be possibly be used as as a parameter
    Variable, Function, list of parameters, callable, or just return a python object

    Args:
        param: Any
        time: Model supplied time
        computed_values: Model supplied computed values
        parameters: Parameters dictionary
        mul_outputs: If param is a list, return the product of its components

    Raises:
        Exception: Raised if param is a Variable with unknown source

    Returns:
        Any: Parameter output
    """

    raise Exception("Catch")

    if isinstance(param, float):
        return param
    elif isinstance(param, GraphObject):
        sources = dict(
            computed_values=computed_values,
            parameters=parameters,
            model_variables={"time": time, "computed_values": computed_values},
        )
        return param.evaluate(**sources)
    elif isinstance(param, Variable):
        if param.source == "parameters":
            return parameters[param.key]
        elif param.source == "computed_values":
            return computed_values[param.key]
        else:
            raise Exception("Unsupported variable source", param, param.source)
    elif isinstance(param, Function):
        sources = dict(
            computed_values=computed_values, parameters=parameters, model_variables={"time": time}
        )
        args, kwargs = build_args(param.args, param.kwargs, sources)
        return param.func(*args, **kwargs)
    elif callable(param):
        return param(time, computed_values)
    elif (isinstance(param, list) or isinstance(param, tuple)) and mul_outputs:
        value = 1.0
        for subparam in param:
            value *= get_model_param_value(subparam, time, computed_values, parameters)
        return value
    else:
        return param


def get_static_param_value(obj: Any, static_graph_values: dict, passthrough=True) -> Any:
    """Get the value of a parameter, or of a function that depends only on parameters,
       or return obj if any other type

    Args:
        obj : The Variable, Function, or Python object
        parameters: Parameters dictionary
        passthrough: If True, will return unknown value types "as-is"; default is to raise TypeError

    Returns:
        The value of the object
    """
    from .param_impl import GraphObjectParameter

    # Might have some nested special classes
    if isinstance(obj, dict):
        return {
            k: get_static_param_value(v, static_graph_values, passthrough) for k, v in obj.items()
        }
    elif isinstance(obj, GraphObjectParameter):
        return static_graph_values[obj._graph_key]
    else:
        if passthrough:
            return obj
        else:
            raise TypeError(obj)



from __future__ import annotations
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from summer2 import CompartmentalModel

from inspect import getfullargspec
from typing import Union, Any

from collections.abc import Iterable
from numbers import Real

import numpy as np

# from pydantic import BaseModel
# from pydantic.main import ModelMetaclass

from summer2.parameters import Parameter, Function, Data
from computegraph import ComputeGraph
from computegraph.utils import expand_nested_dict, is_var

GraphObj = Union[Function, Data]


class ParamStruct:
    pass


class ModelBuilder:
    def __init__(self, params: dict, param_class: type):
        self._params = params
        self._params_expanded = expand_nested_dict(params, include_parents=True)
        self.params = self._pyd_params = param_class(**params)
        label_parameters(self.params, params)

    def set_model(self, model: CompartmentalModel):
        self.model = model
        model.builder = self

    def _get_func_args(self, key):
        return [Parameter(k) for k in [*self._params_expanded[key]]]

    def _get_func_kwargs(self, key):
        return {k: Parameter(k) for k in [*self._params_expanded[key]]}

    def find_key_from_obj(self, obj):
        return find_key_from_obj(obj, self.params, self._params, None)

    def find_obj_from_key(self, key):
        return find_obj_from_key(key, self.params)

    def get_param(self, key: str, create: bool = False):
        """Get a Parameter (computegraph Variable) for the given key
        If this key is not contained in the initial parameters, register it
        as a required additional parameter (or raise an Exception if create is False)

        Args:
            key: Key of the parameter
            create: Add this parameter as an output if required - will raise an exception
                    if False and key is not already present in the input parameters

        Returns:
            computegraph Parameter
        """
        if isinstance(key, str):
            key = key
        elif is_var(key, "parameters"):
            key = key.name
        else:
            key = find_key_from_obj(key, self.params, self._params_expanded)
        if key not in self._params_expanded:
            if create:
                if key not in self.required_outputs:
                    self.required_outputs.add(key)
            else:
                raise KeyError(f"Parameter {key} not found in input parameters")
        return Parameter(key)

    def _get_value(self, key: str):
        """Return the initial parameter value for the given key

        Args:
            key (str): Parameter key
        """
        if key in self._params_expanded:
            return self._params_expanded[key]
        else:
            raise KeyError("Key not found in initial parameters", key)

    def get_mapped_func(self, func: callable, param_obj: ParamStruct, kwargs=None):
        argspec = getfullargspec(func)

        kwargs = {} if kwargs is None else kwargs

        supplied_argkeys = list(kwargs)

        msg = f"Function arguments {argspec} do not match pydantic object {param_obj}"
        assert all(
            [hasattr(param_obj, arg) for arg in argspec.args if arg not in supplied_argkeys]
        ), msg
        base_key = self.find_key_from_obj(param_obj)
        mapped_args = {arg: Parameter(f"{base_key}.{arg}") for arg in argspec.args}
        mapped_args.update(kwargs)
        return Function(func, [], mapped_args)

    def get_jax_runner(self, jit=True, solver=None):
        self.model.finalize()
        run_everything, _, _ = get_full_runner(self, True, solver=solver)
        if jit:
            from jax import jit as _jit

            run_everything = _jit(run_everything)
        return run_everything

    def get_input_parameters(self):
        return self.model.get_input_parameters()

    def get_default_parameters(self) -> dict:
        iparams = self.get_input_parameters()
        oparams = {}
        for p in iparams:
            obj = self.find_obj_from_key(p)
            oparams[p] = obj.value
        return oparams


def label_parameters(pstruct: ParamStruct, pdict: dict, layer: list = None):
    """Walk the parameter dictionary pdict, alongside the ParamStruct,
       setting concrete parameter names for each AbstractParameter encountered

    Args:
        pstruct (ParamStruct): The ParamStruct describing the model
        pdict (dict): The parameter dictionary used as constructor for pstruct
        layer: The current naming layer (used internally, do not set)
    """
    if layer is None:
        layer = []
    for k, v in pdict.items():
        if isinstance(pstruct, dict):
            cur_pstruct = pstruct[k]
        else:
            cur_pstruct = getattr(pstruct, k)
        if isinstance(cur_pstruct, Parameter):
            param_key = ".".join(layer + [k])
            cur_pstruct.set_key(param_key)
        elif isinstance(cur_pstruct, dict):
            label_parameters(cur_pstruct, v, layer + [k])
        elif isinstance(cur_pstruct, ParamStruct):
            assert isinstance(v, dict)
            label_parameters(cur_pstruct, v, layer + [k])


def find_key_from_obj(obj: Any, pydparams: ParamStruct, params: dict, layer=None, is_dict=False):
    if layer is None:
        layer = []
    for k, v in params.items():

        if is_dict:
            cur_pydobj = pydparams[k]
        else:
            cur_pydobj = getattr(pydparams, k)
        if cur_pydobj is obj:
            if isinstance(cur_pydobj, ParamStruct) or isinstance(cur_pydobj, Parameter):
                return ".".join(layer + [k])
            else:
                raise TypeError("Cannot match against type", type(obj))
        if isinstance(cur_pydobj, dict):
            res = find_key_from_obj(obj, cur_pydobj, v, layer + [k], True)
            if res is not None:
                return res
        elif isinstance(cur_pydobj, ParamStruct):
            assert isinstance(v, dict)
            res = find_key_from_obj(obj, cur_pydobj, v, layer + [k])
            if res is not None:
                return res

    if len(layer) == 0:
        raise Exception(f"Unable to match {obj} in parameters dictionary", obj, pydparams, layer)


def find_obj_from_key(key: str, pydparams: ParamStruct) -> Any:
    """Find the matching object within a BaseModel for a given key

    Args:
        key: The expanded parameter key
        pydparams: The full instantiated BaseModel

    Raises:
        TypeError: Raised if uniqueness of object cannot be guaranteed

    Returns:
        The corresponding object within the BaseModel
    """
    cur_obj = pydparams
    for layer in key.split("."):
        if isinstance(cur_obj, ParamStruct):
            cur_obj = getattr(cur_obj, layer)
        elif isinstance(cur_obj, Iterable):
            cur_obj = cur_obj[layer]
        else:
            raise TypeError("Cannot resolve for type", key, cur_obj, type(cur_obj))
    return cur_obj


def get_full_runner(builder, use_jax=False, solver=None):
    if use_jax:
        from summer2.runner.jax.util import get_runner
        from summer2.runner.jax.model_impl import build_run_model

        jrunner = get_runner(builder.model)
        jax_run_func, jax_runner_dict = build_run_model(jrunner, solver=solver)

    model_input_p = builder.model.get_input_parameters()

    def run_everything(parameters=None, **kwargs):

        params_base = builder._params_expanded.copy()
        if parameters is not None:
            params_base.update(parameters)

        model_params = {k: v for k, v in params_base.items() if k in model_input_p}

        if use_jax:
            return jax_run_func(parameters=model_params)
        else:
            builder.model.run(parameters=model_params, **kwargs)
            return builder.model

    if use_jax:
        return run_everything, jax_run_func, jax_runner_dict
    else:
        return run_everything


def is_real(v):
    return isinstance(v, Real)


def parameter_class(constraint_func=is_real, desc: str = None, full_desc: str = None):

    _desc = desc
    _full_desc = full_desc

    class ConcreteParameter(Parameter):

        constraint = constraint_func
        description = _desc
        full_description = _full_desc

        def __init__(self, value):
            super().__init__(key=None)
            self.value = value

        def __repr__(self):
            pkey = self.key or self.description or "param"
            return f"{pkey}({self.value})"

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v):
            if not isinstance(v, Real):
                raise TypeError("Real required")
            if not cls.constraint(v):
                raise ValueError(f"Constraint failed", cls.constraint, v)
            return cls(v)

    return ConcreteParameter


def parameter_array_class(constraint_func=is_real, desc: str = None, full_desc: str = None):

    _desc = desc
    _full_desc = full_desc

    class ConcreteArrayParameter(Parameter):

        constraint = constraint_func
        description = _desc
        full_description = _full_desc

        def __init__(self, value):
            super().__init__(key=None)
            self.value = np.array(value)

        def __repr__(self):
            pkey = self.key or self.description or "param"
            return f"{pkey}({self.value})"

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, v):
            v = np.array(v)
            if not all([cls.constraint(v_i) for v_i in v]):
                raise ValueError(f"Constraint failed", cls.constraint, v)
            return cls(v)

    return ConcreteArrayParameter



"""
This module contains the classes which are used to calculate inter-compartmental flow rates.
As a user of the framework you should not have to use these classes directly.
"""
from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Set

import numpy as np

from summer2.adjust import BaseAdjustment, FlowParam, Multiply
from summer2.compartment import Compartment
from summer2.parameters.param_impl import ModelParameter
from summer2.stratification import Stratification


class WeightType:
    STATIC = "static"
    FUNCTION = "function"


class BaseFlow(ABC):
    """
    :meta private:
    Abstract base class for all flows.
    A flow represents the movement of people into, between or out of compartments.
    """

    name = None
    source = None
    dest = None
    param = None
    adjustments = None
    is_death_flow = False
    tags = None

    def _is_equal(self, flow):
        """For testing"""
        return (
            type(self) is type(flow)
            and self.name == flow.name
            and self.source == flow.source
            and self.dest == flow.dest
            and self.param == flow.param
            and len(self.adjustments) == len(flow.adjustments)
            and all(
                [
                    self.adjustments[i]._is_equal(flow.adjustments[i])
                    for i in range(len(self.adjustments))
                ]
            )
        )

    def is_match(self, name: str, source_strata: dict, dest_strata: dict) -> bool:
        """
        Returns True if the flow matches the given name and strata.
        """
        return (
            self.name == name
            and ((not source_strata) or (not self.source) or self.source.has_strata(source_strata))
            and ((not dest_strata) or (not self.dest) or self.dest.has_strata(dest_strata))
        )

    def get_weight_value(self, time: float, computed_values: dict, parameters: dict = None):
        """
        Returns the flow's 'weight' (i.e. rate to be multiplied by the value of the source
        compartment) at a given time.
        Applies any stratification adjustments to the base parameter.
        """
        # flow_rate = get_model_param_value(self.param, time, computed_values, parameters, True)
        if isinstance(self.param, ModelParameter):
            flow_rate = self.param.get_value(time, computed_values, parameters)
        elif callable(self.param):
            flow_rate = self.param(time, computed_values)
        else:
            flow_rate = self.param
        for adjustment in self.adjustments:
            flow_rate = adjustment.get_new_value(flow_rate, computed_values, time, parameters)

        return flow_rate

    def update_compartment_indices(self, mapping: Dict[str, float]):
        """
        Update index which maps flow compartments to compartment value array.
        """
        if self.source:
            self.source.idx = mapping[self.source]
        if self.dest:
            self.dest.idx = mapping[self.dest]

    @abstractmethod
    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        """
        Returns the net flow value at a given time.
        """
        pass

    @abstractmethod
    def stratify(self, strat: Stratification) -> list:
        """
        Returns a list of new, stratified flows to replace the current flow.
        """
        pass

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return self.__class__(**kwargs)

    @abstractmethod
    def __repr__(self):
        """
        Returns a text representation of the flow.
        """
        pass


class BaseEntryFlow(BaseFlow):
    """
    :meta private:
    A flow where people enter the destination compartment, but there is no source.
    Eg. births, importation.
    """

    _is_birth_flow = False

    def __init__(
        self,
        name: str,
        dest: Compartment,
        param: FlowParam,
        adjustments: List[BaseAdjustment] = None,
        tags: Set[str] = None,
    ):
        assert type(dest) is Compartment
        self.name = name
        self.adjustments = [a for a in (adjustments or []) if a and a.param is not None]
        self.dest = dest
        self.param = param
        tags = tags or set()
        self.tags = tags

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified entry flows to replace the current flow.
        """
        if not self.dest.has_name_in_list(strat.compartments):
            # Flow destination is not stratified, do not stratify this flow.
            return [self]

        new_flows = []
        flow_adjustments = strat.get_flow_adjustment(self)
        is_birth_into_agegroup_flow = self._is_birth_flow and strat.is_ageing()

        msg = "Cannot adjust birth flows into age stratifications."
        assert not (is_birth_into_agegroup_flow and flow_adjustments), msg

        msg = f"Flow {self.name} has missing adjustments for {strat.name} strat."
        assert not (flow_adjustments and set(flow_adjustments.keys()) != set(strat.strata)), msg

        for stratum in strat.strata:
            new_adjustments = [*self.adjustments]
            if is_birth_into_agegroup_flow:
                # Use special rules for birth into age groups.
                if stratum != "0":
                    # Babies get born at age 0, and not at any other age. Skip!
                    continue
            else:
                # Not an ageing stratification, check for user-specified flow adjustments.
                if flow_adjustments:
                    strata_flow_adjustment = flow_adjustments.get(stratum)
                    if strata_flow_adjustment:
                        new_adjustments.append(strata_flow_adjustment)
                else:
                    # No adjustments specified for this flow.
                    # Default to equally dividing entry population between all strata.
                    entry_fraction = 1.0 / len(strat.strata)
                    new_adjustments.append(Multiply(entry_fraction))

            new_dest = self.dest.stratify(strat.name, stratum)
            new_flow = self.copy(
                name=self.name,
                dest=new_dest,
                param=self.param,
                adjustments=new_adjustments,
                tags=self.tags,
            )
            new_flows.append(new_flow)

        return new_flows

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.name}' to {self.dest}>"


class BaseExitFlow(BaseFlow):
    """
    :meta private:
    A flow where people exit the source compartment, but there is no destination
    Eg. deaths, emigration
    """

    def __init__(
        self,
        name: str,
        source: Compartment,
        param: FlowParam,
        adjustments: List[BaseAdjustment] = None,
        tags: Set[str] = None,
    ):
        assert type(source) is Compartment
        self.name = name
        self.adjustments = [a for a in (adjustments or []) if a and a.param is not None]
        self.source = source
        self.param = param
        tags = tags or set()
        self.tags = tags

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified exit flows to replace the current flow.
        """
        if not self.source.has_name_in_list(strat.compartments):
            # Flow source is not stratified, do not stratify this flow.
            return [self]

        flow_adjustments = strat.get_flow_adjustment(self)

        msg = f"Flow {self.name} has missing adjustments for {strat.name} strat."
        assert not (flow_adjustments and set(flow_adjustments.keys()) != set(strat.strata)), msg

        new_flows = []
        for stratum in strat.strata:
            new_adjustments = [*self.adjustments]
            # Check for user-specified flow adjustments.
            if flow_adjustments:
                strata_flow_adjustment = flow_adjustments.get(stratum)
                if strata_flow_adjustment:
                    new_adjustments.append(strata_flow_adjustment)

            new_source = self.source.stratify(strat.name, stratum)
            new_flow = self.copy(
                name=self.name,
                source=new_source,
                param=self.param,
                adjustments=new_adjustments,
                tags=self.tags,
            )
            new_flows.append(new_flow)

        return new_flows

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.name}' from {self.source}>"


class BaseTransitionFlow(BaseFlow):
    """
    :meta private:
    A flow where people move from the source compartment to the destination.
    e.g. infection, recovery, progression of disease.
    """

    def __init__(
        self,
        name: str,
        source: Compartment,
        dest: Compartment,
        param: FlowParam,
        adjustments: List[BaseAdjustment] = None,
        tags: Set[str] = None,
    ):
        assert type(source) is Compartment
        assert type(dest) is Compartment
        self.name = name
        self.adjustments = [a for a in (adjustments or []) if a and a.param is not None]
        self.source = source
        self.dest = dest
        self.param = param
        tags = tags or set()
        self.tags = tags

    def stratify(self, strat: Stratification) -> List[BaseFlow]:
        """
        Returns a list of new, stratified flows to replace the current flow.
        """
        is_source_compartment_stratified = self.source.has_name_in_list(strat.compartments)
        is_dest_compartment_stratified = self.dest.has_name_in_list(strat.compartments)
        if not (is_dest_compartment_stratified or is_source_compartment_stratified):
            # Flow is not stratified, do not stratify this flow.
            return [self]

        new_flows = []
        flow_adjustments = strat.get_flow_adjustment(self)

        msg = f"Flow {self.name} has missing adjustments for {strat.name} strat."
        assert not (flow_adjustments and set(flow_adjustments.keys()) != set(strat.strata)), msg

        for stratum in strat.strata:
            # Find new compartments
            if is_source_compartment_stratified:
                new_source = self.source.stratify(strat.name, stratum)
            else:
                new_source = self.source

            if is_dest_compartment_stratified:
                new_dest = self.dest.stratify(strat.name, stratum)
            else:
                new_dest = self.dest

            # Find flow adjustments to apply to the new stratified flows.

            # There are three scenarios to consider here:
            # - Both the source and destination are stratified.
            # - The flow source has the required stratifications and the destination does not.
            #   For example - people recovering from I -> R with multiple I strata, all with
            #   different recovery rates.
            # - The destination has the required stratifications and the source does not.
            #   For example - people recovering from I -> R with multiple R strata, with different
            #   recovery proportions.
            new_adjustments = [*self.adjustments]

            # Should we apply an adjustment to conserve the flow rate?
            should_apply_conservation_split = (
                # If the destination has been stratified but the source hasn't been.
                (is_dest_compartment_stratified and not is_source_compartment_stratified)
                # Don't conserve flow rates for disease strain stratifications.
                and (not strat.is_strain())
                # Don't do this if the user has specified adjustments.
                and (not flow_adjustments)
            )
            if should_apply_conservation_split:
                # If the source is stratified but not the destination, then we need to account
                # for the resulting fan-out of flows by reducing the flow rate.
                # We don't do this for strains because this effect is already
                # captured by the infectiousness multiplier.
                entry_fraction = 1.0 / len(strat.strata)
                new_adjustments.append(Multiply(entry_fraction))
            elif flow_adjustments:
                # Use user-specified flow adjustments.
                strata_flow_adjustment = flow_adjustments.get(stratum)
                new_adjustments.append(strata_flow_adjustment)

            new_flow = self.copy(
                name=self.name,
                source=new_source,
                dest=new_dest,
                param=self.param,
                adjustments=new_adjustments,
                tags=self.tags,
            )
            new_flows.append(new_flow)

        return new_flows

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.name}' from {self.source} to {self.dest}>"


class CrudeBirthFlow(BaseEntryFlow):
    """
    A flow that calculates births using a 'crude birth rate' method.
    The number of births will be determined by the product of the birth rate and total population.

    Args:
        name: The flow name.
        dest: The destination compartment.
        param: The fraction of the population to be born per timestep.
        adjustments: Adjustments to the flow rate.

    """

    _is_birth_flow = True

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        parameter_value = self.get_weight_value(time, computed_values, parameters)
        total_population = np.sum(compartment_values)
        return parameter_value * total_population


class ReplacementBirthFlow(BaseEntryFlow):
    """
    A flow that calculates births by replacing total deaths.
    The total number of deaths will be supplied by the parameter function.

    Args:
        name: The flow name.
        dest: The destination compartment.
        param: The total number of deaths per timestep.
        adjustments: Adjustments to the flow rate.

    """

    _is_birth_flow = True

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        return self.get_weight_value(time, computed_values, parameters)


class ImportFlow(BaseEntryFlow):
    """
    Calculates importation, where people enter the destination compartment from outside the system.
    The number of people imported per timestep is independent of the source compartment, because
    there isn't a source compartment.

    Args:
        name: The flow name.
        dest: The destination compartment.
        param: The number of people to be imported per timestep.
        adjustments: Adjustments to the flow rate.

    """

    _is_birth_flow = False

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        return self.get_weight_value(time, computed_values, parameters)


class DeathFlow(BaseExitFlow):
    """
    A flow representing deaths.
    Calculated from a fractional death rate.

    Args:
        name: The flow name.
        dest: The destination compartment.
        param: The fraction of the population who die per timestep.
        adjustments: Adjustments to the flow rate.

    """

    is_death_flow = True

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        parameter_value = self.get_weight_value(time, computed_values, parameters)
        population = compartment_values[self.source.idx]
        flow_rate = parameter_value * population
        return flow_rate


class TransitionFlow(BaseTransitionFlow):
    """
    A flow that transfers people from a source to a destination based on
    the population of the source compartment and the fractional flow rate.

    Args:
        name: The flow name.
        source: The source compartment.
        dest: The destination compartment.
        param: The fraction of the source compartment to transfer per timestep.
        adjustments: Adjustments to the flow rate.

    """

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        parameter_value = self.get_weight_value(time, computed_values, parameters)
        population = compartment_values[self.source.idx]
        return parameter_value * population


class AbsoluteFlow(BaseTransitionFlow):
    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        parameter_value = self.get_weight_value(time, computed_values, parameters)
        return parameter_value


class BaseInfectionFlow(BaseTransitionFlow):
    def __init__(
        self,
        name: str,
        source: Compartment,
        dest: Compartment,
        param: FlowParam,
        find_infectious_multiplier: Callable[[Compartment, Compartment], float],
        adjustments: List[BaseAdjustment] = None,
        tags: Set[str] = None,
    ):
        assert type(source) is Compartment
        assert type(dest) is Compartment
        self.name = name
        self.adjustments = [a for a in (adjustments or []) if a and a.param is not None]
        self.source = source
        self.dest = dest
        self.param = param
        self.find_infectious_multiplier = find_infectious_multiplier
        tags = tags or set()
        tags.add("infection")
        self.tags = tags

    def get_net_flow(
        self,
        compartment_values: np.ndarray,
        computed_values: dict,
        time: float,
        parameters: dict = None,
    ) -> float:
        multiplier = self.find_infectious_multiplier(self.source, self.dest)
        parameter_value = self.get_weight_value(time, computed_values, parameters)
        population = compartment_values[self.source.idx]
        return parameter_value * population * multiplier

    def copy(self, **kwargs):
        """
        Creates a modified copy of the flow for stratification.
        """
        return self.__class__(
            **kwargs,
            find_infectious_multiplier=self.find_infectious_multiplier,
        )


class InfectionDensityFlow(BaseInfectionFlow):
    """
    An infection flow that should use the number of infectious people to calculate the force of
    infection factor.
    """


class InfectionFrequencyFlow(BaseInfectionFlow):
    """
    An infection flow that should use the prevalence of infectious people to calculate the force
    of infection factor.
    """



"""Implementation of CompartmentalModel and ModelBackend internals in Jax

This is a mess right now!
"""

from functools import partial
import numpy as np

from jax import jit, numpy as jnp
from summer2.runner.jax import ode
from summer2.runner.jax import solvers

from summer2.adjust import Overwrite

from summer2.runner import ModelBackend

from summer2.solver import SolverType, SolverArgs

from .stratify import get_calculate_initial_pop
from .derived_outputs import build_derived_outputs_runner


def clean_compartments(compartment_values: jnp.ndarray):
    return jnp.where(compartment_values < 0.0, 0.0, compartment_values)


def get_strain_infection_values(
    strain_infectious_values,
    strain_compartment_infectiousness,
    strain_category_indexer,
    mixing_matrix,
    category_populations,
):
    infected_values = strain_infectious_values * strain_compartment_infectiousness
    infectious_populations = jnp.sum(infected_values[strain_category_indexer], axis=-1)
    infection_density = mixing_matrix @ infectious_populations
    category_prevalence = infectious_populations / category_populations
    infection_frequency = mixing_matrix @ category_prevalence

    return {"infection_density": infection_density, "infection_frequency": infection_frequency}


def build_get_infectious_multipliers(runner):
    population_cat_indexer = jnp.array(runner._population_category_indexer)

    # FIXME: We are hardcoding this for frequency only right now
    infect_proc_type = runner._infection_process_type

    if infect_proc_type == "both":
        raise NotImplementedError("No support for mixed infection frequency/density")

    # FIXME: This could desparately use a tidy-up - all the indexing is a nightmare

    def get_infectious_multipliers(
        time, compartment_values, cur_graph_outputs, compartment_infectiousness
    ):

        infection_frequency = {}
        infection_density = {}

        full_multipliers = jnp.ones(len(runner.infectious_flow_indices))

        mixing_matrix = cur_graph_outputs["mixing_matrix"]
        category_populations = compartment_values[population_cat_indexer].sum(axis=1)

        for strain_idx, strain in enumerate(runner.model._disease_strains):

            strain_compartment_infectiousness = compartment_infectiousness[strain]
            strain_infectious_idx = runner._strain_infectious_indexers[strain]
            strain_category_indexer = runner._strain_category_indexers[strain]

            strain_infectious_values = compartment_values[strain_infectious_idx]
            strain_values = get_strain_infection_values(
                strain_infectious_values,
                strain_compartment_infectiousness,
                strain_category_indexer,
                mixing_matrix,
                category_populations,
            )
            infection_frequency[strain] = strain_values["infection_frequency"]
            infection_density[strain] = strain_values["infection_density"]

            if infect_proc_type == "freq":
                strain_ifect = strain_values["infection_frequency"]
            elif infect_proc_type == "dens":
                strain_ifect = strain_values["infection_density"]

            # FIXME: So we produce strain infection values _per category_
            # (ie not all model compartments, not all infectious compartments)
            # We need to rebroadcast these back out the appropriate compartments somehow
            # ... hence the following bit of weird double-dip indexing

            # _infect_strain_lookup_idx is an array of size (num_infectious_comps), whose values
            # are the strain_idx of the strain to whom they belong

            strain_ifectcomp_mask = runner._infect_strain_lookup_idx == strain_idx

            # _infect_cat_lookup_idx is an array of size (num_infectious_comps), whose values
            # are the mixing category to which they belong (and therefore an index into
            # the values returned in strain_ifect above)

            # Get the strain infection values broadcast to num_infectious_comps
            strain_ifect_bcast = strain_ifect[runner._infect_cat_lookup_idx]

            # full_multipliers is the length of the infectious compartments - ie the same as the
            # above mask
            # full_multipliers = full_multipliers.at[strain_ifectcomp_mask].mul(
            #    strain_ifect_bcast[strain_ifectcomp_mask]
            # )

            full_multipliers = full_multipliers.at[strain_ifectcomp_mask].set(
                full_multipliers[strain_ifectcomp_mask] * strain_ifect_bcast[strain_ifectcomp_mask]
            )

        return full_multipliers

    return get_infectious_multipliers


def build_get_flow_weights(runner: ModelBackend):

    m = runner.model

    if "model_variables.time" in m.graph.dag:
        tvkeys = list(m.graph.filter(sources="model_variables.time").dag)
        tv_flow_map = {
            k: m._flow_key_map[k] for k in set(m._flow_key_map).intersection(set(tvkeys))
        }
    else:
        tv_flow_map = {}

    def get_flow_weights(cur_graph_outputs, static_flow_weights):

        flow_weights = jnp.copy(static_flow_weights)

        for k, v in tv_flow_map.items():
            val = cur_graph_outputs[k]
            flow_weights = flow_weights.at[v].set(val)

        return flow_weights

    return get_flow_weights


def _build_calc_computed_values(runner):
    def calc_computed_values(compartment_vals, time, parameters):
        model_variables = {"compartment_values": compartment_vals, "time": time}

        computed_values = runner.computed_values_runner(
            parameters=parameters, model_variables=model_variables
        )

        return computed_values

    return calc_computed_values


def build_get_flow_rates(runner, ts_graph_func):

    # calc_computed_values = build_calc_computed_values(runner)
    get_flow_weights = build_get_flow_weights(runner)

    infect_proc_type = runner._infection_process_type
    if infect_proc_type:
        get_infectious_multipliers = build_get_infectious_multipliers(runner)

    # flow_weights_base = jnp.array(runner.flow_weights)

    population_idx = np.array(runner.population_idx)
    infectious_flow_indices = np.array(runner.infectious_flow_indices)

    def get_flow_rates(compartment_values: jnp.array, time, static_graph_vals, model_data):

        # COULD BE JITTED
        compartment_values = clean_compartments(compartment_values)

        # runner._prepare_time_step(time, compartment_vals)
        sources = {
            "model_variables": {"time": time, "compartment_values": compartment_values},
            "static_inputs": static_graph_vals,
        }

        ts_graph_vals = ts_graph_func(**sources)

        # JITTED
        # computed_values = calc_computed_values(compartment_values, time, parameters)

        # JITTED
        flow_weights = get_flow_weights(ts_graph_vals, model_data["static_flow_weights"])

        populations = compartment_values[population_idx]

        # Update for special cases (population-independent and CrudeBirth)
        if runner._has_non_pop_flows:
            populations = populations.at[runner._non_pop_flow_idx].set(1.0)
        if runner._has_crude_birth:
            populations = populations.at[runner._crude_birth_idx].set(compartment_values.sum())

        flow_rates = flow_weights * populations

        # Calculate infection flows
        # JITTED
        if infect_proc_type:
            infect_mul = get_infectious_multipliers(
                time,
                compartment_values,
                ts_graph_vals,
                model_data["compartment_infectiousness"],
            )
            flow_rates = flow_rates.at[infectious_flow_indices].mul(infect_mul)

            # ReplacementBirthFlow depends on death flows already being calculated; update here
        if runner._has_replacement:
            # Only calculate timestep_deaths if we use replacement, it's expensive...
            _timestep_deaths = flow_rates[runner.death_flow_indices].sum()
            flow_rates = flow_rates.at[runner._replacement_flow_idx].set(_timestep_deaths)

        return flow_rates, ts_graph_vals["computed_values"]

    return get_flow_rates


def build_get_compartment_rates(runner):
    accum_maps = get_accumulation_maps(runner)

    def get_compartment_rates(compartment_values, flow_rates):
        comp_rates = jnp.zeros_like(compartment_values)

        for (flow_src, comp_target) in accum_maps["positive"]:
            comp_rates = comp_rates.at[comp_target].add(flow_rates[flow_src])
        for (flow_src, comp_target) in accum_maps["negative"]:
            comp_rates = comp_rates.at[comp_target].add(-flow_rates[flow_src])

        return comp_rates

    return get_compartment_rates


def build_get_rates(runner, ts_graph_func):
    get_flow_rates = build_get_flow_rates(runner, ts_graph_func)
    get_compartment_rates = build_get_compartment_rates(runner)

    def get_rates(compartment_values, time, static_graph_vals, model_data):
        flow_rates, _ = get_flow_rates(compartment_values, time, static_graph_vals, model_data)
        comp_rates = get_compartment_rates(compartment_values, flow_rates)

        return flow_rates, comp_rates

    return {"get_flow_rates": get_flow_rates, "get_rates": get_rates}


def get_accumulation_maps(runner):
    pos_map = [mflow for mflow in runner._pos_flow_map]
    neg_map = [mflow for mflow in runner._neg_flow_map]

    def peel_flow_map(flow_map):
        targets = []
        src_idx = []
        remainder = []
        for (src_flow, target) in flow_map:
            if target not in targets:
                targets.append(target)
                src_idx.append(src_flow)
            else:
                remainder.append((src_flow, target))
        return np.array(src_idx), np.array(targets), remainder

    def recurse_unpeel(flow_map):
        remainder = flow_map
        full_map = []
        while len(remainder) > 0:
            sources, targets, remainder = peel_flow_map(remainder)
            full_map.append((sources, targets))
        return full_map

    return {"positive": recurse_unpeel(pos_map), "negative": recurse_unpeel(neg_map)}


def build_get_compartment_infectiousness(model):
    """
    Build a Jax function to return the compartment infectiousness (for all compartments),
    of the strain specified by strain
    """

    # This is run during prepare_dynamic
    # i.e. it is done once at the start of a model run, but
    # is parameterized (non-structural)
    def get_compartment_infectiousness(static_graph_values):

        # Find the infectiousness multipliers for each compartment being implemented in the model.
        compartment_infectiousness = jnp.ones(len(model.compartments))

        # Apply infectiousness adjustments
        for strat in model._stratifications:
            for comp_name, adjustments in strat.infectiousness_adjustments.items():
                for stratum, adjustment in adjustments.items():
                    if adjustment:
                        is_overwrite = isinstance(adjustment, Overwrite)
                        adj_value = static_graph_values[adjustment.param._graph_key]
                        adj_comps = model.get_matching_compartments(
                            comp_name, {strat.name: stratum}
                        )
                        for c in adj_comps:
                            if is_overwrite:
                                compartment_infectiousness = compartment_infectiousness.at[
                                    c.idx
                                ].set(adj_value)
                            else:
                                orig_value = compartment_infectiousness[c.idx]
                                compartment_infectiousness = compartment_infectiousness.at[
                                    c.idx
                                ].set(adj_value * orig_value)

        strain_comp_inf = {}

        for strain in model._disease_strains:
            if "strain" in model.stratifications:
                strain_filter = {"strain": strain}
            else:
                strain_filter = {}

            # _Must_ be ordered here
            strain_infect_comps = model.query_compartments(
                strain_filter, tags="infectious", as_idx=True
            )

            strain_comp_inf[strain] = compartment_infectiousness[strain_infect_comps]

        return strain_comp_inf

    return get_compartment_infectiousness


def build_run_model(runner, base_params=None, dyn_params=None, solver=None, solver_args=None):

    if dyn_params is None:
        dyn_params = runner.model.graph.get_input_variables()
    else:
        dyn_params = [
            f"parameters.{p}" if not p.startswith("parameters.") else p for p in dyn_params
        ]

    # dyn_params may contain parameters only used in derived outputs, that do not show up
    # in the main graph.  These need to be filtered out; they will be computed every time,
    # but it's simpler than trying to squash the graphs together
    model_graph_keys = set(runner.model.graph.dag)
    dyn_params = [k for k in dyn_params if k in model_graph_keys]

    # Graph frozen for all non-calibration parameters
    if base_params is None:
        base_params = {}

    source_inputs = {"parameters": base_params}

    ts_vars = runner.model.graph.query("model_variables")

    dyn_params = set(dyn_params).union(set(ts_vars))

    param_frozen_cg = runner.model.graph.freeze(dyn_params, source_inputs)

    # static_cg = param_frozen_cg.filter(exclude=ts_vars)
    # static_graph_func = static_cg.get_callable()(parameters=base_params)

    timestep_cg, static_cg = param_frozen_cg.freeze(ts_vars)

    timestep_graph_func = jit(timestep_cg.get_callable())
    # timestep_graph_func = timestep_cg.get_callable()
    static_graph_func = static_cg.get_callable()

    rates_funcs = build_get_rates(runner, timestep_graph_func)
    get_rates = rates_funcs["get_rates"]
    get_flow_rates = rates_funcs["get_flow_rates"]

    from jax import vmap

    get_flows_for_outputs = vmap(get_flow_rates, in_axes=(0, 0, None, None), out_axes=(0))

    def get_comp_rates(comp_vals, t, static_graph_vals, model_data):
        return get_rates(comp_vals, t, static_graph_vals, model_data)[1]

    if solver is None or solver == SolverType.SOLVE_IVP:
        solver = SolverType.ODE_INT

    if solver == SolverType.ODE_INT:
        if solver_args is None:
            # Some sensible defaults; faster than
            # the odeint defaults,
            # but accurate enough for our tests
            solver_args = SolverArgs.DEFAULT

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return ode.odeint(
                get_comp_rates,
                initial_population,
                times,
                static_graph_vals,
                model_data,
                **solver_args,
            )

    elif solver == SolverType.RUNGE_KUTTA:

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return solvers.rk4(
                get_comp_rates, initial_population, times, static_graph_vals, model_data
            )

    elif solver == SolverType.EULER:

        def get_ode_solution(initial_population, times, static_graph_vals, model_data):
            return solvers.euler(
                get_comp_rates, initial_population, times, static_graph_vals, model_data
            )

    else:
        raise NotImplementedError("Incompatible SolverType for Jax runner", solver)

    times = jnp.array(runner.model.times)

    calc_initial_pop = get_calculate_initial_pop(runner.model)
    get_compartment_infectiousness = build_get_compartment_infectiousness(runner.model)

    do_cg, calc_derived_outputs = build_derived_outputs_runner(runner.model)

    m = runner.model
    if "model_variables.time" in m.graph.dag:
        tv_keys = list(m.graph.filter(sources="model_variables.time").dag)
    else:
        tv_keys = []

    static_flow_map = {k: m._flow_key_map[k] for k in set(m._flow_key_map).difference(set(tv_keys))}

    def run_model(parameters):

        static_graph_vals = static_graph_func(parameters=parameters)
        initial_population = calc_initial_pop(static_graph_vals)

        static_flow_weights = jnp.zeros(len(runner.model._flows))
        for k, v in static_flow_map.items():
            val = static_graph_vals[k]
            static_flow_weights = static_flow_weights.at[v].set(val)

        compartment_infectiousness = get_compartment_infectiousness(static_graph_vals)
        model_data = {
            "compartment_infectiousness": compartment_infectiousness,
            "static_flow_weights": static_flow_weights,
        }

        outputs = get_ode_solution(initial_population, times, static_graph_vals, model_data)

        out_flows, out_cv = get_flows_for_outputs(outputs, times, static_graph_vals, model_data)

        model_variables = {"outputs": outputs, "flows": out_flows, "computed_values": out_cv}

        derived_outputs = calc_derived_outputs(
            parameters=parameters, model_variables=model_variables
        )

        # return {"outputs": outputs, "model_data": model_data}
        return {
            "outputs": outputs,
            "derived_outputs": derived_outputs,
        }  # "model_data": model_data}

    def one_step(parameters: dict = None):

        static_graph_vals = static_graph_func(parameters=parameters)
        initial_population = calc_initial_pop(static_graph_vals)

        static_flow_weights = jnp.zeros(len(runner.model._flows))
        for k, v in static_flow_map.items():
            val = static_graph_vals[k]
            static_flow_weights = static_flow_weights.at[v].set(val)

        compartment_infectiousness = get_compartment_infectiousness(static_graph_vals)
        model_data = {
            "compartment_infectiousness": compartment_infectiousness,
            "static_flow_weights": static_flow_weights,
        }

        flow_rates, comp_rates = get_rates(
            initial_population, runner.model.times[0], static_graph_vals, model_data
        )

        # return {"outputs": outputs, "model_data": model_data}
        return {
            "flow_rates": flow_rates,
            "comp_rates": comp_rates,
            "initial_population": initial_population,
            "static_graph_vals": static_graph_vals,
        }  # "model_data": model_data}

    runner_dict = {
        "get_rates": get_rates,
        "get_flow_rates": get_flow_rates,
        "get_comp_rates": get_comp_rates,
        "calc_initial_pop": calc_initial_pop,
        "get_compartment_infectiousness": get_compartment_infectiousness,
        "get_ode_solution": get_ode_solution,
        "calc_derived_outputs": calc_derived_outputs,
        "timestep_graph_func": timestep_graph_func,
        "timestep_cg": timestep_cg,
        "static_cg": static_cg,
        "static_graph_func": static_graph_func,
        "one_step": one_step,
        "derived_outputs_cg": do_cg,
    }

    return run_model, runner_dict



"""
This module presents classes which are used to define stratifications,
which can be applied to the model.
"""
from typing import Callable, Dict, List, Optional, Union

from numbers import Real

import numpy as np
from computegraph.utils import is_var
from computegraph.types import GraphObject

from summer2.adjust import Multiply, Overwrite, enforce_multiply
from summer2.compartment import Compartment
from summer2.parameters import is_func, get_static_param_value
from summer2.parameters.param_impl import get_modelparameter_from_param

Adjustment = Union[Multiply, Overwrite]
MixingMatrix = Union[np.ndarray, Callable[[float], np.ndarray]]

# Allowable error in requested compartment splits between one and the total requested
COMP_SPLIT_REQUEST_ERROR = 1e-2


class Stratification:
    """
    A generic stratification applied to the compartmental model.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.

    """

    _is_ageing = False
    _is_strain = False

    def __init__(
        self,
        name: str,
        strata: List[str],
        compartments: List[str],
    ):
        self.name = name
        self.strata = list(map(str, strata))
        self.compartments = [Compartment(c) if type(c) is str else c for c in compartments]

        # Split population evenly by default.
        num_strata = len(self.strata)
        self.population_split = {s: 1 / num_strata for s in self.strata}

        # Flows are not adjusted by default.
        self.flow_adjustments = {}
        self.infectiousness_adjustments = {}

        # Store internal fast lookup table for flow adjustments using frozenset
        self._flow_adjustments_fs = {}

        # No heterogeneous mixing matrix by default.
        self.mixing_matrix = None

        # Enable/disable validation (runtime assertions)
        # Not directly user accessable - is toggled by CompartmentalModel
        self._validate = True

    def __repr__(self):
        return f"Stratification: {self.name}"

    def is_ageing(self) -> bool:
        """Returns ``True`` if this stratification represents a set of age groups
        with ageing dynamics"""
        return self._is_ageing

    def is_strain(self) -> bool:
        """Returns ``True`` if this stratification represents a set of disease strains"""
        return self._is_strain

    def set_population_split(self, proportions: Dict[str, float]):
        """
        Sets how the stratification will split the population between the strata.

        Args:
            proportions: A map of proportions to be assigned to each strata.

        The supplied proportions must:

            - Specify all strata
            - All be positive
            - Sum to 1 +/- error defined above

        """
        if all([isinstance(v, Real) for v in proportions.values()]):
            self.validate_population_split(proportions)

        self.population_split = proportions

    def validate_population_split(self, proportions: dict):
        msg = f"All strata must be specified when setting population split: {proportions}"
        assert set(list(proportions.keys())) == set(self.strata), msg
        msg = f"All proportions must be >= 0 when setting population split: {proportions}"
        assert all([v >= 0 for v in proportions.values()]), msg
        msg = f"All proportions sum to 1+/-{COMP_SPLIT_REQUEST_ERROR} when setting \
            population split: {proportions}"
        assert abs(1 - sum(proportions.values())) < COMP_SPLIT_REQUEST_ERROR, msg

    def set_flow_adjustments(
        self,
        flow_name: str,
        adjustments: Dict[str, Adjustment],
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
    ):
        """
        Set an adjustment of a flow to the stratification.
        Adjustments from previous stratifications will be applied before this.
        You can use time-varying functions for infectiousness adjustments.

        It is possible to specify multiple flow adjustments for the same flow in a
        given Stratification.
        In this case, only the last-created applicable adjustment will be applied.

        Args:
            flow_name: The name of the flow to adjust.
            adjustments: A dict of adjustments to apply to the flow.
            source_strata (optional): A whitelist of strata to filter the target flow's
            source compartments.
            dest_strata (optional): A whitelist of strata to filter the target flow's
            destination compartments.

        Example:
            Create an adjustment for the 'recovery' flow based on location::

                strat = Stratification(
                    name="location",
                    strata=["urban", "rural", "alpine"],
                    compartments=["S", "I", "R"]
                )
                strat.set_flow_adjustments("recovery", {
                    "urban": Multiply(1.5),
                    "rural": Multiply(0.8),
                    "alpine": None, # No adjustment
                })

        """
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}
        msg = "You must specify all strata when adding flow adjustments."
        assert set(adjustments.keys()) == set(self.strata), msg

        adjustments = {k: enforce_multiply(v) for k, v in adjustments.items()}

        msg = "All flow adjustments must be Multiply, Overwrite or None."
        assert all(
            [
                type(adj) is Overwrite or type(adj) is Multiply or adj is None
                for adj in adjustments.values()
            ]
        ), msg

        msg = "Cannot add new flow adjustments after stratification has already been applied"
        assert len(self._flow_adjustments_fs) == 0, msg

        if flow_name not in self.flow_adjustments:
            self.flow_adjustments[flow_name] = []

        # FIXME Turn this into class or named tuple or something... we want to know what's going on
        self.flow_adjustments[flow_name].append((adjustments, source_strata, dest_strata))

    def get_flow_adjustment(self, flow) -> dict:
        """
        Note that the loop structure implies that if the user has requested multiple adjustments
        that apply to a single combination of strata (across multiple stratifications), then only
        the last one that is applicable will be used - because the last request will over-write the
        earlier ones in the loop.
        Therefore, the most recently added flow adjustment that matches a given flow will be
        returned.

        """
        flow_adjustments = self.flow_adjustments.get(flow.name, [])
        matching_adjustment = None

        # Cache frozenset lookup
        if flow.name not in self._flow_adjustments_fs:
            self._flow_adjustments_fs[flow.name] = cur_fadj_fs = []
            for adjustment, source_strata, dest_strata in flow_adjustments:
                cur_fadj_fs.append(
                    (
                        adjustment,
                        source_strata,
                        frozenset(source_strata.items()),
                        dest_strata,
                        frozenset(dest_strata.items()),
                    )
                )

        flow_adj_fs = self._flow_adjustments_fs[flow.name]

        # Loop over all the requested adjustments.
        for adjustment, source_strata, _source_strata, dest_strata, _dest_strata in flow_adj_fs:

            # These are expensive assertions and can be disabled in multi-run situations
            if self._validate:
                # For entry flows:
                msg = f"Source strata requested in flow adjustment of {self.name}, but {flow.name} \
                    does not have a source"
                assert not (source_strata and not flow.source), msg

                # For exit flows:
                msg = f"Dest strata requested in flow adjustment of {self.name}, but {flow.name} \
                    does not have a dest"
                assert not (dest_strata and not flow.dest), msg

            # Make sure that the source request applies to this flow because it has all of the
            # requested strata.
            # Note that these can be specified in the current or any previous stratifications.
            is_source_no_match = (
                source_strata and flow.source and not flow.source._has_strata(_source_strata)
            )
            if is_source_no_match:
                continue
            is_dest_no_match = dest_strata and flow.dest and not flow.dest._has_strata(_dest_strata)
            if is_dest_no_match:
                continue

            matching_adjustment = adjustment

        return matching_adjustment

    def add_infectiousness_adjustments(
        self, compartment_name: str, adjustments: Dict[str, Adjustment]
    ):
        """
        Add an adjustment of a compartment's infectiousness to the stratification.
        You cannot currently use time-varying functions for infectiousness adjustments.
        All strata in this stratification must be specified as keys in the adjustments argument,
        if no adjustment required for a stratum, specify None as the value to the request for
        that stratum.

        Args:
            compartment_name: The name of the compartment to adjust.
            adjustments: An dict of adjustments to apply to the compartment.

        Example:
            Create an adjustment for the 'I' compartment based on location::

                strat = Stratification(
                    name="location",
                    strata=["urban", "rural", "alpine"],
                    compartments=["S", "I", "R"]
                )
                strat.add_infectiousness_adjustments("I", {
                    "urban": adjust.Multiply(1.5),
                    "rural": adjust.Multiply(0.8),
                    "alpine": None, # No adjustment
                })

        """
        msg = "You must specify all strata when adding infectiousness adjustments."
        assert set(adjustments.keys()) == set(self.strata), msg

        adjustments = {k: enforce_multiply(v) for k, v in adjustments.items()}

        # for k, v in adjustments.items():
        #    if v is not None:
        #        v.param = get_modelparameter_from_param(v.param)

        msg = "All infectiousness adjustments must be Multiply, Overwrite or None."
        assert all(
            [type(a) is Overwrite or type(a) is Multiply or a is None for a in adjustments.values()]
        ), msg

        msg = f"An infectiousness adjustment for {compartment_name} \
                already exists for strat {self.name}"
        assert compartment_name not in self.infectiousness_adjustments, msg

        # FIXME: This should work for computegraph Functions that are not time dependant,
        # (and fail for those that are), but it's tricky to check...
        # msg = "Cannot use time varying functions for infectiousness adjustments."
        # assert not any([adj.param.is_time_varying() for adj in adjustments.values() if adj]), msg

        self.infectiousness_adjustments[compartment_name] = adjustments

    def set_mixing_matrix(self, mixing_matrix: MixingMatrix):
        """
        Sets the mixing matrix for the model.
        Note that this must apply to all compartments, although this is checked at runtime rather
        than here.
        """
        msg = "Strain stratifications cannot have a mixing matrix."
        assert not self.is_strain(), msg

        # msg = "Mixing matrix must be a NumPy array, or return a NumPy array."
        # assert (
        #    (isinstance(mixing_matrix, np.ndarray))
        #    or isinstance(mixing_matrix, GraphObject)
        #    or is_func(mixing_matrix)
        # ), msg

        # if isinstance(mixing_matrix, np.ndarray):
        #    num_strata = len(self.strata)
        #    msg = f"Mixing matrix must have both {num_strata} rows and {num_strata} columns."
        #    assert mixing_matrix.shape == (num_strata, num_strata), msg

        self.mixing_matrix = mixing_matrix

    def _stratify_compartments(self, comps: List[Compartment]) -> List[Compartment]:
        """
        Stratify the model compartments into sub-compartments, based on the strata names,
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartments.
        """
        strat_base_indices = []
        passthrough_base_indices = []
        passthrough_target_indices = []
        stratum_target_indices = {s: [] for s in self.strata}

        new_comps = []
        idx = 0
        for base_idx, old_comp in enumerate(comps):
            should_stratify = old_comp.has_name_in_list(self.compartments)
            if should_stratify:
                strat_base_indices.append(base_idx)
                for stratum in self.strata:
                    new_comp = old_comp.stratify(self.name, stratum)
                    new_comp.idx = idx
                    new_comps.append(new_comp)
                    stratum_target_indices[stratum].append(idx)
                    idx += 1
            else:
                passthrough_base_indices.append(base_idx)
                passthrough_target_indices.append(idx)
                old_comp.idx = idx
                new_comps.append(old_comp)
                idx += 1

        self._strat_base_indices = np.array(strat_base_indices, dtype=int)
        self._passthrough_base_indices = np.array(passthrough_base_indices, dtype=int)
        self._passthrough_target_indices = np.array(passthrough_target_indices, dtype=int)
        self._stratum_target_indices = {
            k: np.array(v, dtype=int) for k, v in stratum_target_indices.items()
        }
        self._new_size = idx

        return new_comps

    def _stratify_compartment_values(
        self, comps: List[Compartment], comp_values: np.ndarray, static_graph_values: dict = None
    ) -> np.ndarray:
        """
        Stratify the model compartments into sub-compartments, based on the strata names provided.
        Split the population according to the provided proportions.
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartment values.
        """
        assert len(comps) == len(comp_values)
        new_comp_values = []

        if static_graph_values:
            population_split = get_static_param_value(self.population_split, static_graph_values)
        else:
            population_split = self.population_split

        for idx in range(len(comp_values)):
            should_stratify = comps[idx].has_name_in_list(self.compartments)
            if should_stratify:
                for stratum in self.strata:
                    new_value = comp_values[idx] * population_split[stratum]
                    new_comp_values.append(new_value)
            else:
                new_comp_values.append(comp_values[idx])

        return np.array(new_comp_values)


class AgeStratification(Stratification):
    """
    A stratification that represents a set of age groups with ageing dynamics.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.


    Strata must be a list of strings or integers that represent all ages. For example, the strata
    [0, 10, 20, 30] will be interpreted as the age groups 0-9, 10-19, 20-29, 30+ respectively.

    Using this stratification will automatically add a set of ageing flows to the model,
    where the exit rate is proportional to the age group's time span. For example, in the 0-9
    strata, there will be a flow created where 10% of occupants "age up" into the 10-19 strata
    each year.
    **Critically**, this feature assumes that each of the model's time steps represents a year.

    """

    _is_ageing = True

    def __init__(
        self,
        name: str,
        strata: List[str],
        compartments: List[str],
    ):
        try:
            _strata = sorted(map(int, strata))
        except Exception:
            raise AssertionError("Strata must be in an int-compatible format")

        assert _strata[0] == 0, "First age strata must be 0"

        _strata = map(str, _strata)
        super().__init__(name, _strata, compartments)


class StrainStratification(Stratification):
    """
    A stratification that represents a set of disease strains.

    Args:
        name: The name of the stratification.
        strata: The new strata that will be created.
        compartments: The compartments which will be stratified.


    Each requested stratum will be interpreted as a different strain of the disease being modelled.
    This will mean that the force of infection calculations will consider each strain separately.

    Strain stratifications cannot use a mixing matrix

    """

    _is_strain = True



"""
Optimized 'hot' functions used by CompartmentalModel and its runners.
"""

from jax import jit, numpy as np
from functools import partial

# Use Numba to speed up the calculation of the population.
@jit
def find_sum(compartment_values: np.ndarray) -> float:
    return compartment_values.sum()


@jit
def accumulate_positive_flow_contributions(
    flow_rates: np.ndarray,
    comp_rates: np.ndarray,
    pos_flow_map: np.ndarray,
):
    """
    Fast accumulator for summing positive flow rates into their effects on compartments

    Args:
        flow_rates (np.ndarray): Flow rates to be accumulated
        comp_rates (np.ndarray): Output array of compartment rates
        pos_flow_map (np.ndarray): Array of src (flow), target (compartment) indices
    """
    comp_rates = comp_rates.copy()
    for src, target in pos_flow_map:
        comp_rates = comp_rates.at[target].add(flow_rates[src])
    return comp_rates


@jit
def accumulate_negative_flow_contributions(
    flow_rates: np.ndarray,
    comp_rates: np.ndarray,
    neg_flow_map: np.ndarray,
):
    """Fast accumulator for summing negative flow rates into their effects on compartments

    Args:
        flow_rates (np.ndarray): Flow rates to be accumulated
        comp_rates (np.ndarray): Output array of compartment rates
        neg_flow_map (np.ndarray): Array of src (flow), target (compartment) indices
    """
    comp_rates = comp_rates.copy()
    for src, target in neg_flow_map:
        comp_rates = comp_rates.at[target].add(-flow_rates[src])
    return comp_rates


@jit
def get_strain_infection_values(
    strain_infectious_values,
    strain_compartment_infectiousness,
    strain_category_indexer,
    mixing_matrix,
    category_populations,
):
    infected_values = strain_infectious_values * strain_compartment_infectiousness
    infectious_populations = np.sum(infected_values[strain_category_indexer], axis=1)
    infection_density = mixing_matrix @ infectious_populations
    category_prevalence = infectious_populations / category_populations
    infection_frequency = mixing_matrix @ category_prevalence

    return infection_density, infection_frequency



"""
Calculation of derived outputs.

A derived output is an additional output that the user has requested,
which can be calculated (or "derived") using the model results, which are the:

    - evaluation times
    - compartment sizes for each time
    - flow rates for each time

"""
import logging
from typing import Callable, Dict, List, Optional

import networkx
import numpy as np

import pandas as pd

from summer2.compartment import Compartment
from summer2.flows import BaseFlow
from summer2.utils import get_scenario_start_index

from summer2.parameters import build_args

logger = logging.getLogger()


class DerivedOutputRequest:
    FLOW = "flow"
    COMPARTMENT = "comp"
    AGGREGATE = "agg"
    CUMULATIVE = "cum"
    FUNCTION = "func"
    COMPUTED_VALUE = "computed_value"
    PARAM_FUNCTION = "param_func"


def calculate_derived_outputs(
    requests: List[dict],
    graph: networkx.DiGraph,
    outputs: np.ndarray,
    times: np.ndarray,
    timestep: float,
    flows: List[BaseFlow],
    compartments: List[Compartment],
    get_flow_rates: Callable[[np.ndarray, float], np.ndarray],
    model,
    whitelist: Optional[List[str]],
    baseline=None,
    idx_cache=None,
    parameters: dict = None,
) -> Dict[str, np.ndarray]:
    """
    Calculates all requested derived outputs from the calculated compartment sizes.

    Args:
        requests: Descriptions of the new outputs we should create.
        graph: A DAG describing how the requests depend on each other.
        outputs: The compartmental model outputs - compartment sizes over time.
        times: The times that the outputs correspond to.
        timestep: The timestep used to generate the model times.
        flows: The flows used by the model.
        compartments: The compartments used by the model.
        get_flow_rates: A function that gets the model flow rates for a given state and time.
        whitelist: An optional subset of requests to evaluate.
        baseline: Optional CompartmentalModel object to be used as a reference

    Returns:
        Dict[str, np.ndarray]: The timeseries results for each requested output.

    """
    assert outputs is not None, "Cannot calculate derived outputs: model has not been run."
    error_msg = "Cannot calculate derived outputs: dependency graph has cycles."
    assert networkx.is_directed_acyclic_graph(graph), error_msg
    graph = graph.copy()  # We're going to mutate the graph so copy it first.

    if whitelist:
        # Only calculate the required outputs and their dependencies, ignore everything else.
        required_nodes = set()
        for name in whitelist:
            # Find a list of the output required and its dependencies.
            output_dependencies = networkx.dfs_tree(graph.reverse(), source=name).reverse()
            required_nodes = required_nodes.union(output_dependencies.nodes)

        # Remove any nodes that aren't required from the graph.
        nodes = list(graph.nodes)
        for node in nodes:
            if not node in required_nodes:
                graph.remove_node(node)

    derived_outputs = {}
    outputs_to_delete_after = []

    # If we have a baseline for comparison, get some basic information re offsets
    if baseline:
        baseline_start_index = get_scenario_start_index(baseline.times, times[0])

    # Calculate all flow rates and store in `flow_values` so that we can fulfil flow rate requests.
    # We need to do this here because some solvers do not necessarily evaluate all timesteps.
    flow_values = np.zeros((len(times), len(flows)))

    # These are 'extra' values computed by requested processes, and need to be tracked separately
    computed_values = []

    # FIXME: Another question for Matt - has my changes to the time requests stuffed this up?
    # Because the timestep for the last time interval can now be different from the earlier ones.
    # So do we need to assert that the duration is an exact multiple of the timestep?
    # Could cause silent problems, because presumably we have previously been specifying durations as multiples of the timestep.
    for time_idx, time in enumerate(times):
        # Flow rates are instantaneous; we need to provide an integrated value over timestep
        flow_rates_t, computed_values_t = get_flow_rates(outputs[time_idx], time)
        flow_values[time_idx, :] = flow_rates_t * timestep
        # Collect these as lists then build DataFrames afterwards
        computed_values.append(computed_values_t)

    # Collate list values into DataFrames
    computed_values = pd.DataFrame(
        columns=model.get_computed_value_keys(), data=computed_values, index=times
    )

    # Convert tracked flow values into a matrix where the 1st dimension is flow type, 2nd is time
    flow_values = np.array(flow_values).T

    # Calculate all the outputs in the correct order so that each output has its dependencies fulfilled.
    for name in networkx.topological_sort(graph):
        request = requests[name]
        request_type = request["request_type"]
        output = np.zeros(times.shape)

        if not request["save_results"]:
            # Delete the results of this output once the calcs are done.
            outputs_to_delete_after.append(name)

        if request_type == DerivedOutputRequest.FLOW:
            # User wants to track a set of flow rates over time.
            output, idx_cache = _get_flow_output(
                request, name, times, flows, flow_values, idx_cache
            )
        elif request_type == DerivedOutputRequest.COMPARTMENT:
            # User wants to track a set of compartment sizes over time.
            output, idx_cache = _get_compartment_output(
                request, name, outputs, compartments, idx_cache
            )
        elif request_type == DerivedOutputRequest.AGGREGATE:
            # User wants to track the sum of a set of outputs over time.
            output = _get_aggregate_output(request, derived_outputs)
        elif request_type == DerivedOutputRequest.CUMULATIVE:
            # User wants to track cumulative value of an output over time.
            if baseline:
                baseline_offset = baseline.derived_outputs[name][baseline_start_index]
            else:
                baseline_offset = None
            output = _get_cumulative_output(request, name, times, derived_outputs, baseline_offset)
        elif request_type == DerivedOutputRequest.FUNCTION:
            # User wants to track the results of a function of other outputs over time.
            output = _get_func_output(request, derived_outputs)
        elif request_type == DerivedOutputRequest.PARAM_FUNCTION:
            output = _get_param_func_output(request, derived_outputs, computed_values, parameters)
        # FIXME DerivedValue and InputValue should probably be combined
        elif request_type == DerivedOutputRequest.COMPUTED_VALUE:
            output = _get_computed_value_output(request, computed_values)

        derived_outputs[name] = output

    # Delete any intermediate outputs that we don't want to save.
    for name in outputs_to_delete_after:
        del derived_outputs[name]

    return derived_outputs, idx_cache


def _get_flow_output(request, name, times, flows, flow_values, idx_cache=None):
    this_flow_values = np.zeros_like(times)

    if not idx_cache:
        idx_cache = {}

    if not name in idx_cache:
        idx_cache[name] = []
        for flow_idx, flow in enumerate(flows):
            is_matching_flow = (
                flow.name == request["flow_name"]
                and ((not flow.source) or flow.source.has_strata(request["source_strata"]))
                and ((not flow.dest) or flow.dest.has_strata(request["dest_strata"]))
            )
            if is_matching_flow:
                idx_cache[name].append(flow_idx)

    for flow_idx in idx_cache[name]:
        this_flow_values += flow_values[flow_idx]

    use_raw_results = request["raw_results"]
    if use_raw_results:
        # Use interpolated flow rates with no post-processing.
        return this_flow_values, idx_cache
    else:
        # Set the "flow value" at time `t` to be a midpoint approximation of the integrated flow
        # bewteen `t-1` and 't'
        # This is done so that we can estimate the number of people moving between compartments
        # using tracked flow rates.
        midpoint_output = np.zeros(times.shape)
        # First point just uses existing value since no 't=-1' step is available
        # Client using these outputs will typically the first value
        midpoint_output[0] = this_flow_values[0]
        midpoint_output[1:] = (this_flow_values[1:] + this_flow_values[:-1]) * 0.5
        return midpoint_output, idx_cache


def _get_compartment_output(request, name, outputs, compartments, idx_cache=None):
    if not idx_cache:
        idx_cache = {}

    if not name in idx_cache:
        req_compartments = request["compartments"]
        strata = request["strata"]
        comps = ((i, c) for i, c in enumerate(compartments) if c.has_name_in_list(req_compartments))
        idx_cache[name] = [i for i, c in comps if c.is_match(c.name, strata)]

    idxs = idx_cache[name]
    return outputs[:, idxs].sum(axis=1), idx_cache


def _get_aggregate_output(request, derived_outputs):
    source_names = request["sources"]
    return sum([derived_outputs[s] for s in source_names])


def _get_cumulative_output(request, name, times, derived_outputs, baseline_offset=None):
    source_name = request["source"]
    start_time = request["start_time"]
    max_time = times.max()
    if start_time and start_time > max_time:
        # Handle case where the derived output starts accumulating after the last model timestep.
        msg = f"Cumulative output '{name}' start time {start_time} is greater than max model time {max_time}, defaulting to {max_time}"
        logger.warn(msg)
        start_time = max_time

    if start_time is None:
        output = np.cumsum(derived_outputs[source_name])
    else:
        assert start_time in times, f"Start time {start_time} not in times for '{name}'"
        start_idx = np.where(times == start_time)[0][0]
        output = np.zeros_like(times)
        output[start_idx:] = np.cumsum(derived_outputs[source_name][start_idx:])

    if baseline_offset:
        output += baseline_offset

    return output


def _get_func_output(request, derived_outputs):
    func = request["func"]
    source_names = request["sources"]
    inputs = [derived_outputs[s] for s in source_names]
    return func(*inputs)


def _get_param_func_output(request, derived_outputs, computed_values, parameters):
    mfunc = request["func"]
    sources = dict(
        derived_outputs=derived_outputs, parameters=parameters, computed_values=computed_values
    )
    args, kwargs = build_args(mfunc.args, mfunc.kwargs, sources)
    return mfunc.func(*args, **kwargs)


def _get_computed_value_output(request, computed_values):
    name = request["name"]
    return computed_values[name].to_numpy(dtype=float)


def get_scenario_start_index(base_times, start_time):
    """
    Returns the index of the closest time step that is at, or before the scenario start time.
    """
    assert (
        base_times[0] <= start_time
    ), f"Scenario start time {start_time} is before baseline has started"
    indices_after_start_index = [idx for idx, time in enumerate(base_times) if time > start_time]
    if not indices_after_start_index:
        raise ValueError(f"Scenario start time {start_time} is set after the baseline time range")

    index_after_start_index = min(indices_after_start_index)
    start_index = max([0, index_after_start_index - 1])
    return start_index



# flake8: noqa

from jax import lax, numpy as jnp


def euler(get_comp_rates, initial_population, times, model_params, model_data, timescale=1):

    times = jnp.linspace(times[0], times[-1], len(times) * timescale)
    timestep = 1.0 / timescale

    out_vals = jnp.empty((len(times), len(initial_population)))
    out_vals = out_vals.at[0].set(initial_population)

    max_i = len(times) - 1

    def body(state):
        i, out_vals = state
        t = times[i]
        comp_vals = out_vals[i]
        comp_rates = get_comp_rates(comp_vals, t, model_params, model_data)
        comp_vals = comp_vals + comp_rates * timestep
        out_vals = out_vals.at[i + 1].set(comp_vals)
        return i + 1, out_vals

    def cond(state):
        i, _ = state
        return i < max_i

    _, final = lax.while_loop(cond, body, (0, out_vals))
    return final


def rk4(get_comp_rates, initial_population, times, model_params, model_data):

    # times = jnp.linspace(times[0],times[-1],len(times) * timescale)
    # timestep = 1.0/timescale
    timestep = 1.0

    out_vals = jnp.empty((len(times), len(initial_population)))
    out_vals = out_vals.at[0].set(initial_population)

    max_i = len(times) - 1

    def body(state):
        i, out_vals = state
        t = times[i]
        comp_vals = out_vals[i]

        k1 = get_comp_rates(comp_vals, t, model_params, model_data)
        k2 = get_comp_rates(comp_vals + k1 / 2, t + timestep / 2, model_params, model_data)
        k3 = get_comp_rates(comp_vals + k2 / 2, t + timestep / 2, model_params, model_data)
        k4 = get_comp_rates(comp_vals + k3, t + timestep, model_params, model_data)
        comp_vals = comp_vals + (1 / 6) * (k1 + 2 * k2 + 2 * k3 + k4)

        out_vals = out_vals.at[i + 1].set(comp_vals)
        return i + 1, out_vals

    def cond(state):
        i, _ = state
        return i < max_i

    _, final = lax.while_loop(cond, body, (0, out_vals))
    return final



"""Utilities for population redistribution
"""

from typing import NamedTuple

import numpy as np

from summer2.parameters import get_static_param_value


class CompartmentGroup(NamedTuple):
    """Hashable compartment groupings"""

    name: str
    strata: frozenset


def get_unique_strat_groups(comps, strat):
    """
    Return a set of unique name,strata for the given set of compartments,
    which differ only by the strata for 'strat', ie the set of sets
    whose children are specified by 'strat'
    """
    unique_strat_groups = set()
    for c in comps:
        cur_strata = c.strata.copy()
        cur_strata.pop(strat)
        unique_strat_groups.add(CompartmentGroup(c.name, frozenset(cur_strata.items())))
    return unique_strat_groups


def filter_by_strata(comps, strata):
    _strata = frozenset(strata.items())
    return [c for c in comps if c._has_strata(_strata)]


def get_rebalanced_population(
    model,
    population: np.ndarray,
    static_graph_values: dict,
    strat: str,
    dest_filter: dict,
    proportions: dict,
):
    """Adjust the initial population to redistribute the population for a particular
    stratification, over a subset of some other strata

    Args:
        model: The CompartmentalModel to adjust
        strat (str): The stratification to redistribute over
        dest_filter (dict): Subset of (other) strata to filter the split by
        proportions (dict): Proportions of new split (must have all strata specified)

    """

    msg = f"No stratification {strat} found in model"
    assert strat in [s.name for s in model._stratifications], msg

    model_strat = [s for s in model._stratifications if s.name == strat][0]

    proportions = get_static_param_value(proportions, static_graph_values)

    msg = "All strata must be specified in proportions"
    assert set(model_strat.strata) == set(proportions), msg

    msg = "Proportions must sum to 1.0"
    np.testing.assert_allclose(sum(proportions.values()), 1.0, err_msg=msg)

    strat_comps = [c for c in model.compartments if strat in c.strata]
    # Filter by only the subset we're setting in split_map
    strat_comps = filter_by_strata(strat_comps, dest_filter)

    usg = get_unique_strat_groups(strat_comps, strat)

    out_population = population.copy()

    for g in usg:
        mcomps = model._get_matching_compartments(g.name, g.strata)
        idx = np.array([c.idx for c in mcomps])
        total = population[idx].sum()
        for c in mcomps:
            k = c.strata[strat]
            target_prop = proportions[k]
            out_population = out_population.at[c.idx].set(total * target_prop)

    return out_population


def calculate_initial_population(model, parameters=None) -> np.ndarray:
    """
    Called to recalculate the initial population from either fixed dictionary, or a dict
    supplied as a parameter
    """
    if parameters is None:
        parameters = {}
    # FIXME:
    # Work in progress; correctly recalculates non-parameterized
    # populations, but does not include population rebalances etc
    distribution = model._init_pop_dist
    initial_population = np.zeros_like(model._original_compartment_names, dtype=float)

    # if is_var(distribution, "parameters"):
    #    distribution = self.parameters[distribution.name]
    # elif isinstance(distribution, Function) or isinstance(distribution, ModelParameter):
    #    distribution = get_static_param_value(distribution, parameters)

    if isinstance(distribution, dict):
        for idx, comp in enumerate(model._original_compartment_names):
            pop = get_static_param_value(distribution[comp.name], parameters, True)
            assert pop >= 0, f"Population for {comp.name} cannot be negative: {pop}"
            initial_population[idx] = pop

        comps = model._original_compartment_names

        for action in model.tracker.all_actions:
            if action.action_type == "stratify":
                strat = action.kwargs["strat"]
                # for strat in self.model._stratifications:
                # Stratify compartments, split according to split_proportions
                prev_compartment_names = comps  # copy.copy(self.compartments)
                comps = strat._stratify_compartments(comps)
                initial_population = strat._stratify_compartment_values(
                    prev_compartment_names, initial_population, parameters
                )
            elif action.action_type == "adjust_pop_split":
                initial_population = get_rebalanced_population(
                    model, initial_population, parameters, **action.kwargs
                )
        return initial_population
    else:
        raise TypeError(
            "Initial population distribution must be a dict",
            distribution,
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from summer2 import CompartmentalModel

from functools import partial
from typing import Tuple, List, Iterable, Any
from numbers import Real

from computegraph.types import GraphObject, Data, Function, Variable
from computegraph.utils import defer, invert_dict, assign
from computegraph import ComputeGraph
from computegraph.jaxify import get_modules

fnp = get_modules()["numpy"]

import numpy as np

from summer2.parameters.params import (
    build_args,
    ComputedValuesDict,
    Time,
)


class ModelParameter:
    def get_value(self, time: float, computed_values: dict, parameters: dict):
        raise NotImplementedError

    def __eq__(self, other):
        return hash(other) == hash(self)

    def is_time_varying(self):
        return False


class _ComputedValueParameter(ModelParameter):
    def __init__(self, name):
        self.name = name

    def get_value(self, time: float, computed_values: dict, parameters: dict):
        return computed_values[self.name]

    def __hash__(self):
        return hash((self.name, "computed_values"))

    def __repr__(self):
        return f"ComputedValue: {self.name}"

    def is_time_varying(self):
        return True


class _GraphFunction(ModelParameter):
    def __init__(self, func):
        self.func = func

    def get_value(self, time: float, computed_values: dict, parameters: dict):
        sources = dict(
            computed_values=computed_values, parameters=parameters, model_variables={"time": time}
        )
        args, kwargs = build_args(self.func.args, self.func.kwargs, sources)
        return self.func.func(*args, **kwargs)

    def __hash__(self):
        return hash(self.func)

    def __repr__(self):
        return f"GraphFunction: {self.func}"

    def is_time_varying(self):
        has_time_var = Time in self.func.args or Time in self.func.kwargs.values()
        has_computed_value = any([isinstance(arg, ComputedValue) for arg in self.func.args])
        has_computed_value = has_computed_value or any(
            [isinstance(arg, ComputedValue) for arg in self.func.kwargs.values()]
        )

        return has_time_var or has_computed_value


class GraphObjectParameter(ModelParameter):
    def __init__(self, obj):
        self.obj = obj

    def get_value(self, time: float, computed_values: dict, parameters: dict):
        sources = dict(
            computed_values=computed_values,
            parameters=parameters,
            model_variables={"time": time, "computed_values": computed_values},
        )
        return self.obj.evaluate(**sources)

    def __hash__(self):
        return hash(self.obj)

    def __eq__(self, other):
        return self.obj == other.obj

    def is_time_varying(self):
        return True


def get_modelparameter_from_param(param, allow_any=False) -> Tuple[ModelParameter, List[str]]:
    """Create a ModelParameter subclass object that can be called inside a model loop
    (ie time, computed_values and parameters are available)
    These are mostly flow params and adjustments
    Return this object, as well as a list of keys of all parameters found during its
    construction

    Args:
        param: Anything that might resolve to a single float
        allow_any: Whether to allow any generic data and return as Data

    Raises:
        TypeError: Raised if the input param cannot be resolved

    Returns:
        The resolved ModelParameter, along with any parameter keys encountered in the resolution
    """
    if isinstance(param, ModelParameter):
        # We've already transformed this parameter
        return param
    elif isinstance(param, GraphObject):
        return GraphObjectParameter(param)
    elif isinstance(param, Real):
        return GraphObjectParameter(Data(param))
    elif callable(param):
        return GraphObjectParameter(defer(param)(Time, ComputedValuesDict))
    else:
        if allow_any:
            return GraphObjectParameter(Data(param))
        else:
            raise TypeError(f"Unsupported model parameter type {type(param)}", param)


def get_reparameterized_dict(d):
    return {k: get_modelparameter_from_param(v) for k, v in d.items()}


def map_flow_keys(m: CompartmentalModel) -> dict:

    from summer2.adjust import Overwrite

    realised_flows = {}

    for i, f in enumerate(m._flows):
        full_flow = [f.param.obj]
        for a in f.adjustments:
            if isinstance(a, Overwrite):
                full_flow = [a.param.obj]
            else:
                full_flow.append(a.param.obj)
        out_func = full_flow[0]
        for fparam in full_flow[1:]:
            out_func = out_func * fparam
        realised_flows[i] = GraphObjectParameter(out_func)

    return realised_flows

def register_object_key(obj_table, obj, base_name, unique=False):
    if isinstance(obj, dict):
        for k, v in obj.items():
            register_object_key(obj_table, v, f"{base_name}_{k}")
    elif isinstance(obj, GraphObjectParameter):
        if obj.obj not in obj_table:
            if unique:
                name = base_name
                if name in obj_table.values():
                    raise KeyError("Object with name {name} already exists")
            else:
                if base_name in obj_table.values():
                    name = f"{base_name}_{len(obj_table)}"
                else:
                    name = base_name
            obj_table[obj.obj] = name
        obj_key = obj_table[obj.obj]
        obj._graph_key = obj_key
        return obj_key
    else:
        raise TypeError(obj, base_name)

def finalize_parameters(model):
    """Called as part of model.finalize
    This ensures all parameters (and function calls) have concrete computegraph.Variable
    realisations.  We can scan all possible parameterized sites, and replace params
    with appropriate types, as well as collecting all referenced named parameters

    Args:
        model: The CompartmentalModel to finalize
    """

    obj_table = {}

    register_obj_key = partial(register_object_key, obj_table)

    # Flow parameters and adjustments
    for f in model._flows:
        f.param = get_modelparameter_from_param(f.param)

        for adj in f.adjustments:
            if adj is not None:
                adj.param = get_modelparameter_from_param(adj.param)

    realised_flows = map_flow_keys(model)

    all_flow_keys = {}

    for i, f in enumerate(model._flows):
        fpkey = register_obj_key(realised_flows[i], f"{f.name}_rate")
        f._graph_key = fpkey
        if fpkey not in all_flow_keys:
            all_flow_keys[fpkey] = []
        key_store = all_flow_keys[fpkey]
        key_store.append(i)

    model._flow_key_map = {k: fnp.array(v, dtype=int) for k, v in all_flow_keys.items()}

    # Initial population
    if not hasattr(model, "_init_pop_dist"):
        raise Exception("Model initial population must be set before finalizing")

    model._init_pop_dist = get_reparameterized_dict(model._init_pop_dist)
    register_obj_key(model._init_pop_dist, "init_pop_dist")

    # Keep track of matrices so we can Kron them together later
    mixing_matrices = []
    # Stratifications - population split, infectiousness adjustments, mixing matrix
    for s in model._stratifications:
        s.population_split = get_reparameterized_dict(s.population_split)
        register_obj_key(s.population_split, f"{s.name}_pop_split")

        for comp, adjustments in s.infectiousness_adjustments.items():
            for stratum, adjustment in adjustments.items():
                if adjustment is not None:
                    param = get_modelparameter_from_param(adjustment.param)
                    adjustment.param = param
                    register_obj_key(adjustment.param, f"{s.name}_iadj_{comp}_{stratum}", True)

        if s.mixing_matrix is not None:
            param = get_modelparameter_from_param(s.mixing_matrix, True)
            s.mixing_matrix = param
            matrix_key = f"{s.name}_mixing_matrix"
            register_obj_key(s.mixing_matrix, matrix_key, True)
            mixing_matrices.append(param.obj)

    if len(mixing_matrices) == 0:
        param = get_modelparameter_from_param(Data(fnp.array([[1.0]])))
        register_obj_key(param, "mixing_matrix", True)
        model.mixing_matrix = param
    elif len(mixing_matrices) == 1:
        mm = mixing_matrices[0]
        param = get_modelparameter_from_param(defer(assign)(mm))
        register_obj_key(param, "mixing_matrix", True)
        model.mixing_matrix = param
    else:

        def compute_final_matrix(base_matrix, *args):
            cur_matrix = base_matrix
            for m in args:
                cur_matrix = fnp.kron(cur_matrix, m)
            return cur_matrix

        final_mat_func = defer(compute_final_matrix)(*mixing_matrices)
        param = get_modelparameter_from_param(final_mat_func)
        model.mixing_matrix = param
        register_obj_key(param, "mixing_matrix", True)

    for k, v in model._computed_values_graph_dict.items():
        name = f"computed_values.{k}"
        register_obj_key(GraphObjectParameter(v), name, True)

    # Capture computed values in dictionary so that
    # 'old-style' summer functions (t,cv) can use them
    def capture_kwargs(*args, **kwargs):
        return kwargs

    cv_func = Function(capture_kwargs, kwargs=model._computed_values_graph_dict)
    cv_func.node_name = "gather_cv"
    register_obj_key(GraphObjectParameter(cv_func), "computed_values", True)

    do_table = {}
    register_do_obj_key = partial(register_object_key, do_table)

    # Derived outputs
    for k, v in model._derived_output_requests.items():
        req_type = v["request_type"]
        if req_type == "param_func":
            name = f"derived_outputs.{k}"
            register_do_obj_key(GraphObjectParameter(v["func"]), name, True)

    do_graph = invert_dict(do_table)
    model._do_tracker_graph = ComputeGraph(do_graph, validate_keys=False)

    model_graph = invert_dict(obj_table)

    model.graph = ComputeGraph(model_graph, validate_keys=False)

    # Computed values
    # cv_graph = {}

    # for k, v in model._computed_values_graph_dict.items():
    #    param, pkeys = concretize_function_args(v, builder)
    #    cv_graph[k] = param
    #    all_params += pkeys

    # model._computed_values_graph_dict = cv_graph

    # Derived outputs
    # for k, v in model._derived_output_requests.items():
    #    req_type = v["request_type"]
    #    if req_type == "param_func":
    #        param, pkeys = concretize_function_args(v["func"], builder)
    #        v["func"] = param
    #        all_params += pkeys



"""
This module contains a class used to represent compartments.
"""
from typing import Dict, Optional


class Compartment:
    """
    A single compartment in the compartmental model.
    Each compartment is defined by its name and stratum/strata if stratified.
    A compartment does not store the number of occupants - this data is tracked elsewhere in ``CompartmentalModel``.

    Args:
        name: The compartment's name
        strata: A mapping which defines this compartment's strata for each stratification.

    Example:
        Create a Compartment with no stratifications::

            comp = Compartment("susceptible")

        Create a Compartment with age and location stratifications::

            comp = Compartment(
                name="susceptible"
                strata={
                    "age": "young",
                    "location": "rural",
                }
            )

    """

    def __init__(
        self,
        name: str,
        strata: Optional[Dict[str, str]] = None,
        idx: Optional[int] = None,
        tags: Optional[list] = None,
    ):
        assert type(name) is str, "Name must be a string, not %s." % type(name)
        self.name = name
        self.strata = strata or {}
        self._str = self.serialize()
        self.idx = idx
        self._strata = frozenset(self.strata.items())
        self.tags = tags or []

    def is_match(self, name: str, strata: dict) -> bool:
        """
        Determines whether this compartment matches the supplied name and strata.
        A partial strata match, where this compartment has more strata than are supplied returns True.
        """
        return name == self.name and self.has_strata(strata)

    def has_strata(self, strata: dict) -> bool:
        # Use frozenset for performance - this is _very_ significant
        # It is equivalent to this code:
        # return all([self.has_stratum(k, v) for k, v in strata.items()])
        return frozenset(strata.items()).issubset(self._strata)

    def _has_strata(self, strata: frozenset) -> bool:
        # As above but takes frozenset directly as argument
        return strata.issubset(self._strata)

    def has_stratum(self, stratification: str, stratum: str) -> bool:
        return self.strata.get(stratification) == stratum

    def has_name(self, comp) -> bool:
        """
        Returns True if this compartment has the same root name as another.
        """
        # FIXME: Don't understand why these two alternatives are needed.
        if type(comp) is str:
            return self.name == comp
        else:
            return self.name == comp.name

    def has_name_in_list(self, comps: list) -> bool:
        """
        Returns True if this compartment has the same root name as any compartment in the provided list.
        """
        return any(self.has_name(c) for c in comps)

    def stratify(self, stratification_name: str, stratum_name: str):
        """
        Returns a copy of the Compartment with a new stratification applied to it.

        Args:
            stratification_name: The name (or purpose) of the new stratification.
            stratum_name: The stratum to be assigned to the new Compartment.
        Returns:
            Compartment: The new, stratified compartment.

        """
        new_strata = {**self.strata, stratification_name: stratum_name}
        return Compartment(name=self.name, strata=new_strata, tags=self.tags)

    def serialize(self) -> str:
        """Returns a string representation of the compartment"""
        strata_strings = [f"{k}_{v}" for k, v in self.strata.items()]
        return "X".join([self.name, *strata_strings])

    @staticmethod
    def deserialize(s: str):
        """Converts a string into a Compartment"""
        parts = s.split("X")
        name = parts[0]
        strata = {}

        # FIXME: Does this mean that stratum names should never have underscores in them?
        # ... or more generally are there still strings with specific behaviours that we should be careful of?
        for strat in parts[1:]:
            stratum_parts = strat.split("_")
            s_name = stratum_parts[0]
            s_value = "_".join(stratum_parts[1:])
            strata[s_name] = s_value

        return Compartment(name, strata)

    def __str__(self) -> str:
        return self._str

    def __repr__(self) -> str:
        return self._str

    def __eq__(self, obj):
        """
        Equality operator override
        """
        is_same_as_str = type(obj) is str and obj == self._str
        is_same_as_comp = type(obj) is Compartment and obj._str == self._str
        return is_same_as_str or is_same_as_comp

    def __hash__(self):
        """
        Hash operator override
        """
        return hash(self._str)



from typing import Tuple

import numpy as np
from jax import numpy as jnp

from summer2 import Compartment
from summer2.adjust import Overwrite
import summer2.flows as flows

from summer2.population import get_rebalanced_population
from summer2.utils import clean_compartment_values

class ModelBackend:
    """
    An optimized, but less accessible model runner.
    """

    def __init__(self, model):
        # Compartmental model
        self.model = model

        # Tracks total deaths per timestep for death-replacement birth flows
        self._timestep_deaths = None

        # Set our initial parameters to an empty dict - this is really just to appease tests
        self.parameters = {}

    def prepare_structural(self):
        # FIXME: Redundant
        self._iter_non_function_flows = [(i, f) for i, f in enumerate(self.model._flows)]

        self._build_compartment_category_map()

        self.infectious_flow_indices = np.array(
            [i for i, f in self._iter_non_function_flows if isinstance(f, flows.BaseInfectionFlow)],
            dtype=int,
        )
        self.death_flow_indices = np.array(
            [i for i, f in self._iter_non_function_flows if f.is_death_flow], dtype=int
        )

        # Include dummy values in population_idx to account for Entry flows
        non_func_pops = np.array(
            [f.source.idx if f.source else 0 for i, f in self._iter_non_function_flows], dtype=int
        )

        self.population_idx = non_func_pops

        # Store indices of flows that are not population dependent
        self._non_pop_flow_idx = np.array(
            [
                i
                for i, f in self._iter_non_function_flows
                if (type(f) in (flows.ReplacementBirthFlow, flows.ImportFlow, flows.AbsoluteFlow))
            ],
            dtype=int,
        )
        self._has_non_pop_flows = bool(len(self._non_pop_flow_idx))

        # Crude birth flows use population sum rather than a compartment; store indices here
        self._crude_birth_idx = np.array(
            [i for i, f in self._iter_non_function_flows if type(f) == flows.CrudeBirthFlow],
            dtype=int,
        )
        self._has_crude_birth = bool(len(self._crude_birth_idx))

        self._has_replacement = False
        # Replacement flows must be calculated after death flows, store indices here
        for i, f in self._iter_non_function_flows:
            if type(f) == flows.ReplacementBirthFlow:
                self._has_replacement = True
                self._replacement_flow_idx = i

        self._precompute_flow_maps()
        self._build_infectious_multipliers_lookup()

    def prepare_static_params(self, parameters: dict, dyn_params: list = None):
        """Do all precomputation here"""

        parameters = parameters or {}
        self.parameters = parameters

        #
        source_inputs = {"parameters": parameters}

        # Can replace this with calibration parameters later
        if dyn_params is None:
            dyn_vars = self.model.graph.get_input_variables()
        else:
            dyn_vars = dyn_params

        # Graph frozen for all non-calibration parameters
        self.param_frozen_cg = self.model.graph.freeze(dyn_vars, source_inputs)

        # Query model_variables (ie time-varying sources)
        ts_vars = self.param_frozen_cg.query("model_variables")

        # Subgraph that does not contain model_variables (ie anything time varying)
        self.static_cg = self.param_frozen_cg.filter(exclude=ts_vars)
        self.run_graph_static = self.static_cg.get_callable()

    def prepare_dynamic(self, param_updates: dict = None):
        """Do all precomputation here"""

        param_updates = param_updates or {}

        self.parameters.update(param_updates)

        #
        source_inputs = {"parameters": self.parameters}

        # Query model_variables (ie time-varying sources)
        ts_vars = self.param_frozen_cg.query("model_variables")

        self._graph_values_static = self.run_graph_static(parameters=self.parameters)

        # Subgraph whose only dynamic inputs are model_variables
        self.timestep_cg = self.param_frozen_cg.freeze(ts_vars, source_inputs)

        self.run_graph_ts = self.timestep_cg.get_callable()

        # Must be run after self._graph_values_static is available
        self.calculate_initial_population()

        self._compartment_infectiousness = self._get_compartment_infectiousness()

        # Initialize arrays for infectious multipliers
        # FIXME Put this in its own function, this is getting messy again
        num_cats = self.num_categories
        num_strains = len(self.model._disease_strains)
        self._infection_density = {}
        self._infection_frequency = {}
        self._infection_density_flat = np.empty((num_strains, num_cats), dtype=float)
        self._infection_frequency_flat = np.empty((num_strains, num_cats), dtype=float)

    def prepare_to_run(self, parameters: dict = None):
        # FIXME:
        # This is now only used by tests, and should never called in any production code
        if not hasattr(self.model, "_init_pop_dist"):
            self.model.set_initial_population({}, force=True)
        self.model.finalize()
        self.prepare_structural()
        self.prepare_static_params(parameters)
        self.prepare_dynamic({})

    def _build_compartment_category_map(self):
        # Create a matrix that tracks which categories each compartment is in.
        self.num_categories = ncats = len(self.model._mixing_categories)

        # Array that maps compartments to their mixing category
        self._category_lookup = np.empty(len(self.model.compartments), dtype=int)

        all_cat_idx = []
        for i, category in enumerate(self.model._mixing_categories):
            cat_idx = []
            for j, comp in enumerate(self.model.compartments):
                if all(comp.has_stratum(k, v) for k, v in category.items()):
                    cat_idx.append(j)
                    # self._category_matrix[i][j] = 1
                    self._category_lookup[j] = i
            all_cat_idx.append(np.array(cat_idx, dtype=int))

        # Indexer for whole population (compartment_values), that will return a
        # num_cats * (num_comps_per_cat) array that can be summed for category populations
        self._population_category_indexer = pop_cat_idx = np.stack(all_cat_idx)

        # Array (per-strain) of compartment indices for that strain's infectious compartments
        self._strain_infectious_indexers = {}
        # Array (per-strain) of len (strain infectious comps) that maps each
        # strain infectious compartment to its mixing category
        self._strain_category_indexers = {}
        #
        for strain in self.model._disease_strains:
            strain_filter = {"strain": strain} if "strain" in self.model.stratifications else {}
            strain_infectious_comps = self.model.query_compartments(
                strain_filter, tags="infectious", as_idx=True
            )
            self._strain_infectious_indexers[strain] = jnp.array(strain_infectious_comps)

            # Function that tests each element of an ndarray to see if is contained within the
            # current strain infectious compartments
            vcat = np.vectorize(lambda c: c in strain_infectious_comps)
            strain_cat_idx = pop_cat_idx[vcat(pop_cat_idx)]
            # Localize the compartment indices to be within strain infectious comps,
            # rather than the global compartment lookup
            # We do this by creating an inverse global map (but only fill in the compartment
            # indices we care about)
            # It's just a bit easier to read that messing around with dictionaries etc
            tlookup = np.empty(len(self.model.compartments), dtype=int)
            tlookup[strain_infectious_comps] = range(len(strain_infectious_comps))
            strain_cat_idx = tlookup[strain_cat_idx]

            # Ensure this is a 2d array
            # This necessary where there is only one compartment for each category
            strain_cat_idx = strain_cat_idx.reshape((ncats, int(strain_cat_idx.size / ncats)))
            self._strain_category_indexers[strain] = jnp.array(strain_cat_idx)

    def _get_compartment_infectiousness(self):
        """
        Returns an array of relative infectiousness weights for the supplied strain
        The returned array is of size num_infectious_comps_for_strain, and is ordered
        (ie can be multiplied directly against these infectious compartments)
        """
        # Find the infectiousness multipliers for each compartment being implemented in the model.
        # Start from assumption that each compartment is not infectious.
        compartment_infectiousness = np.ones(self.model.initial_population.shape)
        # Set all infectious compartments to be equally infectious.

        for strat in self.model._stratifications:
            for comp_name, adjustments in strat.infectiousness_adjustments.items():
                for stratum, adjustment in adjustments.items():
                    if adjustment:
                        is_overwrite = isinstance(adjustment, Overwrite)
                        adj_value = self._graph_values_static[adjustment.param._graph_key]
                        adj_comps = self.model.get_matching_compartments(
                            comp_name, {strat.name: stratum}
                        )
                        for c in adj_comps:
                            if is_overwrite:
                                compartment_infectiousness[c.idx] = adj_value
                            else:
                                orig_value = compartment_infectiousness[c.idx]
                                compartment_infectiousness[c.idx] = adj_value * orig_value

        strain_comp_inf = {}

        for strain in self.model._disease_strains:
            if "strain" in self.model.stratifications:
                strain_filter = {"strain": strain}
            else:
                strain_filter = {}

            # _Must_ be ordered here
            strain_infect_comps = self.model.query_compartments(
                strain_filter, tags="infectious", as_idx=True
            )

            strain_comp_inf[strain] = compartment_infectiousness[strain_infect_comps]

        return strain_comp_inf

    def _precompute_flow_maps(self):
        """Build fast-access arrays of flow indices"""
        f_pos_map = []
        f_neg_map = []
        for i, f in self._iter_non_function_flows:
            if f.source:
                f_neg_map.append((i, f.source.idx))
            if f.dest:
                f_pos_map.append((i, f.dest.idx))

        self._pos_flow_map = np.array(f_pos_map, dtype=int)
        self._neg_flow_map = np.array(f_neg_map, dtype=int)

    def _prepare_time_step(self, time: float, compartment_values: np.ndarray):
        """
        Pre-timestep setup. This should be run before `_get_rates`.
        Here we set up any stateful updates that need to happen before we get the flow rates.
        """

        # Some flows (e.g birth replacement) expect this value to be defined
        self._timestep_deaths = 0.0

        model_variables = {
            "time": time,
            "compartment_values": compartment_values,
            "computed_values": {},
        }

        self._graph_values_timestep = self.run_graph_ts(
            parameters=self.parameters, model_variables=model_variables
        )

        # Find the effective infectious population for the force of infection (FoI) calculations.
        self._mixing_matrix = mixing_matrix = self._graph_values_timestep["mixing_matrix"]
        self._calculate_strain_infection_values(compartment_values, mixing_matrix)

    def _get_mixing_matrix(self, t):
        # FIXME: Only used by tests, should never be called in real code
        self.prepare_to_run()
        self._prepare_time_step(t, self.model.initial_population)
        return self._mixing_matrix

    def get_flow_weights(self):
        """Collate weights for all model flows at the current timestep"""

        flow_weights = jnp.zeros(len(self.model._flows))

        for k, v in self.model._flow_key_map.items():
            flow_weights = flow_weights.at[v].set(self._graph_values_timestep[k])

        return flow_weights

    def _get_infectious_multipliers(self) -> np.ndarray:
        """Get multipliers for all infectious flows

        Returns:
            np.ndarray: Array of infectiousness multipliers
        """
        # We can use a fast lookup version if we have only one type of infectious multiplier
        # (eg only frequency, not mixed freq and density)
        if self._infection_frequency_only:
            return self._infection_frequency_flat[
                self._infect_strain_lookup_idx, self._infect_cat_lookup_idx
            ]
        elif self._infection_density_only:
            return self._infection_density_flat[
                self._infect_strain_lookup_idx, self._infect_cat_lookup_idx
            ]

        multipliers = np.empty(len(self.infectious_flow_indices))
        for i, idx in enumerate(self.infectious_flow_indices):
            f = self.model._flows[idx]
            multipliers[i] = f.find_infectious_multiplier(f.source, f.dest)
        return multipliers

    def _build_infectious_multipliers_lookup(self):
        """Get multipliers for all infectious flows

        These are used by _get_infectious_multipliers_flat (currently experimental)

        Returns:
            np.ndarray: Array of infectiousness multipliers
        """
        lookups = []

        has_freq = False
        has_dens = False

        for i, idx in enumerate(self.infectious_flow_indices):
            f = self.model._flows[idx]
            if isinstance(f, flows.InfectionFrequencyFlow):
                has_freq = True
            elif isinstance(f, flows.InfectionDensityFlow):
                has_dens = True
            cat_idx, strain = self._get_infection_multiplier_indices(f.source, f.dest)
            strain_idx = self.model._disease_strains.index(strain)
            lookups.append([strain_idx, cat_idx])
        full_table = np.array(lookups, dtype=int)
        self._full_table = full_table.reshape(len(self.infectious_flow_indices), 2)
        self._infect_strain_lookup_idx = self._full_table[:, 0].flatten()
        self._infect_cat_lookup_idx = self._full_table[:, 1].flatten()

        self._infection_frequency_only = False
        self._infection_density_only = False

        self._infection_process_type = None

        if has_freq:
            if has_dens:
                self._infection_process_type = "both"
            else:
                self._infection_process_type = "freq"
                self._infection_frequency_only = True
        elif has_dens:
            self._infection_density_only = True
            self._infection_process_type = "dens"

    def _get_force_idx(self, source: Compartment):
        """
        Returns the index of the source compartment in the infection multiplier vector.
        """
        return self._category_lookup[source.idx]

    def _get_timestep_deaths(self, *args, **kwargs) -> float:
        assert self._timestep_deaths is not None, "Total deaths has not been set."
        return self._timestep_deaths

    def _get_infection_multiplier_indices(
        self, source: Compartment, dest: Compartment
    ) -> Tuple[str, int]:
        """Return indices for infection frequency lookups"""
        idx = self._get_force_idx(source)
        strain = dest.strata.get("strain", self.model._DEFAULT_DISEASE_STRAIN)
        return idx, strain

    def _get_infection_frequency_multiplier(self, source: Compartment, dest: Compartment) -> float:
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection frequency' calculation.
        """
        idx, strain = self._get_infection_multiplier_indices(source, dest)
        return self._infection_frequency[strain][idx]

    def _get_infection_density_multiplier(self, source: Compartment, dest: Compartment):
        """
        Get force of infection multiplier for a given compartment,
        using the 'infection density' calculation.
        """
        idx, strain = self._get_infection_multiplier_indices(source, dest)
        return self._infection_density[strain][idx]

    def _calculate_strain_infection_values(
        self, compartment_values: np.ndarray, mixing_matrix: np.ndarray
    ):
        # Calculate total number of people per category (for FoI).
        # A vector with size (num_cats).
        from summer2.runner.jax.model_impl import get_strain_infection_values

        self._category_populations = compartment_values[self._population_category_indexer].sum(
            axis=1
        )

        for strain_idx, strain in enumerate(self.model._disease_strains):

            strain_compartment_infectiousness = self._compartment_infectiousness[strain]
            strain_infectious_idx = self._strain_infectious_indexers[strain]
            strain_category_indexer = self._strain_category_indexers[strain]

            strain_infectious_values = compartment_values[strain_infectious_idx]

            

            infection_density, infection_frequency = get_strain_infection_values(
                strain_infectious_values,
                strain_compartment_infectiousness,
                strain_category_indexer,
                mixing_matrix,
                self._category_populations,
            )

            self._infection_density[strain] = infection_density
            self._infection_density_flat[strain_idx] = infection_density

            self._infection_frequency[strain] = infection_frequency
            self._infection_frequency_flat[strain_idx] = infection_frequency

    def _get_flow_rates(self, compartment_vals: np.ndarray, time: float) -> np.ndarray:
        """Get current flow rates, equivalent to calling get_net_flow on all (non-function) flows

        Args:
            comp_vals (np.ndarray): Compartment values
            time (float): Time in model.times coordinates

        Returns:
            np.ndarray: Array of all (non-function) flow rates
        """

        self._prepare_time_step(time, compartment_vals)

        computed_values = self._graph_values_timestep["computed_values"]

        flow_weights = self.get_flow_weights()

        # These will be filled in afterwards
        populations = compartment_vals[self.population_idx]
        # Update for special cases (population-independent and CrudeBirth)
        if self._has_non_pop_flows:
            populations[self._non_pop_flow_idx] = 1.0
        if self._has_crude_birth:
            populations[self._crude_birth_idx] = find_sum(compartment_vals)

        flow_rates = flow_weights * populations

        # Calculate infection flows
        infect_mul = self._get_infectious_multipliers()
        flow_rates = flow_rates.at[self.infectious_flow_indices].mul(infect_mul)

        # ReplacementBirthFlow depends on death flows already being calculated; update here
        if self._has_replacement:
            # Only calculate timestep_deaths if we use replacement, it's expensive...
            self._timestep_deaths = flow_rates[self.death_flow_indices].sum()
            flow_rates[self._replacement_flow_idx] = self._timestep_deaths

        return flow_rates, computed_values

    def _get_rates(self, comp_vals: np.ndarray, time: float) -> Tuple[np.ndarray, np.ndarray]:
        """Calculates inter-compartmental flow rates for a given state and time, as well
        as the updated compartment values once these rate deltas have been applied

        Args:
            comp_vals (np.ndarray): The current state of the model compartments
                                    (ie. number of people)
            time (float): Time in model.times coordinates

        Returns:
            Tuple[np.ndarray, np.ndarray]: (comp_rates, flow_rates) where
                comp_rates is the rate of change of compartments, and
                flow_rates is the contribution of each flow to compartment rate of change
        """

        comp_rates = np.zeros(len(comp_vals))
        flow_rates, _ = self._get_flow_rates(comp_vals, time)

        if self._pos_flow_map.size > 0:
            comp_rates = accumulate_positive_flow_contributions(
                flow_rates, comp_rates, self._pos_flow_map
            )

        if self._neg_flow_map.size > 0:
            comp_rates = accumulate_negative_flow_contributions(
                flow_rates, comp_rates, self._neg_flow_map
            )

        return comp_rates, flow_rates

    def get_compartment_rates(self, compartment_values: np.ndarray, time: float):
        """
        Interface for the ODE solver: this function is passed to solve_ode func and defines the
        dynamics of the model.
        Returns the rate of change of the compartment values for a given state and time.
        """
        comp_vals = clean_compartment_values(compartment_values)
        comp_rates, _ = self._get_rates(comp_vals, time)
        return comp_rates

    def get_flow_rates(self, compartment_values: np.ndarray, time: float):
        """
        Returns the contribution of each flow to compartment rate of change for a given state and
        time.
        """
        comp_vals = clean_compartment_values(compartment_values)
        return self._get_flow_rates(comp_vals, time)

    def calculate_initial_population(self) -> np.ndarray:
        """
        Called to recalculate the initial population from either fixed dictionary, or a dict
        supplied as a parameter
        """
        # FIXME:
        # Work in progress; correctly recalculates non-parameterized
        # populations, but does not include population rebalances etc
        model = self.model

        distribution = model._init_pop_dist
        initial_population = np.zeros_like(model._original_compartment_names, dtype=float)

        # if is_var(distribution, "parameters"):
        #    distribution = self.parameters[distribution.name]
        # elif isinstance(distribution, Function) or isinstance(distribution, ModelParameter):
        #    distribution = get_static_param_value(distribution, parameters)

        if isinstance(distribution, dict):
            for idx, comp in enumerate(model._original_compartment_names):
                pop_key = distribution[comp.name]._graph_key
                pop = self._graph_values_static[pop_key]
                assert pop >= 0, f"Population for {comp.name} cannot be negative: {pop}"
                initial_population[idx] = pop

            comps = model._original_compartment_names

            for action in model.tracker.all_actions:
                if action.action_type == "stratify":
                    strat = action.kwargs["strat"]
                    # for strat in self.model._stratifications:
                    # Stratify compartments, split according to split_proportions
                    prev_compartment_names = comps  # copy.copy(self.compartments)
                    comps = strat._stratify_compartments(comps)
                    initial_population = strat._stratify_compartment_values(
                        prev_compartment_names, initial_population, self._graph_values_static
                    )
                elif action.action_type == "adjust_pop_split":
                    initial_population = get_rebalanced_population(
                        model, initial_population, self._graph_values_static, **action.kwargs
                    )
        else:
            raise TypeError(
                "Initial population distribution must be a dict",
                distribution,
            )
        model.initial_population = initial_population
